#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.shared.view.interfaces module

This module defines interfaces related to views management.
"""

from pyams_content.shared.view import IViewQuery
from pyams_content.shared.view.interfaces.query import IViewQueryFilterExtension, IViewQueryParamsExtension, \
    IViewUserQuery

__docformat__ = 'restructuredtext'


class IEsViewQuery(IViewQuery):
    """Elasticsearch view query marker interface

    This is the base interface of view query.
    """

    def get_post_filters(self, context, request=None, **kwargs):
        """Get static view query post-filters params

        These filters are defined as query parameters, but are applied only after
        aggregates calculation.
        """


class IEsViewQueryParamsExtension(IViewQueryParamsExtension):
    """Elasticsearch view query params extension

    This interface is used to register custom adapters which are defined to get
    Elasticsearch query extra parameters.
    """


class IEsViewUserQuery(IViewUserQuery):
    """Elasticsearch view user query interface"""


class IEsViewQueryPostFilterExtension(IViewQueryParamsExtension):
    """Elasticsearch view query post-filter extension

    This interface is used to register custom adapters which are defined to get
    Elasticsearch post-filter extra parameters; these filters are used to
    filter results after aggregates calculation.
    """


class IEsViewUserPostFilter(IViewUserQuery):
    """Elasticsearch view user post-filter interface"""


class IEsViewQueryResultsFilterExtension(IViewQueryFilterExtension):
    """Elasticsearch view query filter extension

    This interface is used to register adapters which are defined to
    filter results of Elasticsearch query. So unlike params extensions,
    """



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

__docformat__ = 'restructuredtext'

import transaction
from elasticsearch.exceptions import NotFoundError, TransportError

from pyams_content_es.interfaces import IContentIndexerUtility, IDocumentIndexTarget, \
    LOGGER
from pyams_elastic.include import client_from_config
from pyams_layer.skin import apply_skin
from pyams_site.site import site_factory
from pyams_utils.finder import find_objects_providing
from pyams_utils.registry import query_utility, set_local_registry
from pyams_zmi.skin import AdminSkin
from pyams_zmi.utils import get_object_hint, get_object_label


class Args:
    """Args converter class"""

    def __init__(self, args, kwargs):
        self.args = args
        self.kwargs = kwargs

    def __getattr__(self, item):
        """Argument value getter"""
        return self.get(item)

    def get(self, item, default=None):
        """Argument value getter with default"""
        return self.kwargs.get(item, getattr(self.args, item, default))


def index_site(request, autocommit=True, cmd_args=None, **kwargs):
    """Update Elasticsearch index with all database contents"""
    args = Args(cmd_args, kwargs)
    application = site_factory(request)
    if application is not None:
        try:
            set_local_registry(application.getSiteManager())
            indexer = query_utility(IContentIndexerUtility)
            if indexer is None:
                print("Can't find content indexer utility! Aborting...")
            else:

                from pyams_content.shared.common.interfaces import IWfSharedContent

                if args.root is not None:
                    for path in filter(bool, args.root.split('/')):
                        application = application[path]

                apply_skin(request, AdminSkin)
                client = client_from_config(request.registry.settings,
                                            timeout=args.timeout,
                                            use_transaction=False)
                for document in find_objects_providing(application, IDocumentIndexTarget):
                    try:
                        if args.verbose:
                            print(f"Indexing {get_object_hint(document, request)}: "
                                  f"{get_object_label(document, request)} ({document.id})",
                                  end='')
                        if (args.include or args.exclude) and \
                                IWfSharedContent.providedBy(document):
                            if (args.include and document.content_type not in args.include) or \
                                    (args.exclude and document.content_type in args.exclude):
                                if args.verbose:
                                    print(": ignored")
                                else:
                                    print(".", end=' ')
                                continue
                        if args.check:
                            try:
                                client.get(document)
                            except NotFoundError:
                                pass
                            else:
                                if args.verbose:
                                    print(": found")
                                else:
                                    print('-', end=' ')
                                continue
                        try:
                            client.index_object(document)
                        except TransportError:
                            if args.verbose:
                                print("   > TransportError: can't index document")
                            else:
                                print(f"(!! {document.id} !!)", end=' ')
                        else:
                            if args.verbose:
                                print(": index OK")
                            else:
                                print(f"(+{document.id})", end=' ')
                        if autocommit:
                            transaction.commit()
                    except:
                        LOGGER.exception("Error indexing document:")
        finally:
            set_local_registry(None)
        if autocommit:
            transaction.commit()
    return application



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.component.extfile module

This module defines adapters which are used to handle external file indexation.
"""

__docformat__ = 'restructuredtext'

import base64

from pyams_content.component.association.interfaces import IAssociationContainer, IAssociationItem
from pyams_content.component.extfile.interfaces import IBaseExtFile
from pyams_content.component.paragraph.interfaces import IParagraphContainer, \
    IParagraphContainerTarget
from pyams_content_es.interfaces import IDocumentIndexInfo
from pyams_utils.adapter import adapter_config
from pyams_workflow.interfaces import IWorkflow, IWorkflowState


@adapter_config(name='extfile',
                required=IParagraphContainerTarget,
                provides=IDocumentIndexInfo)
def paragraph_container_extfile_index_info(context):
    """Paragraph container external file indexation info"""

    workflow_state = None
    workflow = IWorkflow(context, None)
    if workflow is not None:
        workflow_state = IWorkflowState(context, None)

    # extract paragraphs attachments
    extfiles = []
    attachments = []
    for paragraph in IParagraphContainer(context).get_visible_paragraphs():
        associations = IAssociationContainer(paragraph, {})
        for extfile in associations.values():
            if not IAssociationItem(extfile).visible:
                continue
            if not (IBaseExtFile.providedBy(extfile) and extfile.data):
                continue
            extfiles.append({
                'title': extfile.title,
                'description': extfile.description
            })
            # don't index attachments for contents which are not published
            if workflow_state and (workflow_state.state not in workflow.visible_states):
                continue
            for lang, data in extfile.data.items():
                content_type = data.content_type
                if isinstance(content_type, bytes):
                    content_type = content_type.decode()
                if content_type.startswith('image/') or \
                        content_type.startswith('audio/') or \
                        content_type.startswith('video/'):
                    continue
                attachments.append({
                    'content_type': content_type,
                    'name': data.filename,
                    'language': lang,
                    'content': base64.encodebytes(data.data).decode().replace('\n', '')
                })
    result = {
        'extfile': extfiles
    }
    if attachments:
        result.update({
            '__pipeline__': 'attachment',
            'attachments': attachments
        })
    return result



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.utility module

"""

__docformat__ = 'restructuredtext'

from persistent import Persistent
from zope.container.contained import Contained
from zope.intid import IIntIds
from zope.schema.fieldproperty import FieldProperty

from pyams_content_es.interfaces import IContentIndexerUtility, INDEXER_AUTH_KEY, \
    INDEXER_HANDLER_KEY, IQuickSearchSettings, IUserSearchSettings, LOGGER, QUICK_SEARCH_SETTINGS_KEY, \
    USER_SEARCH_SETTINGS_KEY
from pyams_utils.adapter import adapter_config, get_annotation_adapter
from pyams_utils.factory import factory_config
from pyams_utils.registry import get_pyramid_registry, get_utility
from pyams_utils.transaction import TransactionClient, transactional
from pyams_zmq.socket import zmq_response, zmq_socket


class IndexUpdateHandler(TransactionClient):
    """Content indexer transactional client class"""

    def __init__(self, indexer):
        super().__init__()
        self.indexer = indexer

    @transactional
    def update_index(self, action, document, **kwargs):
        """Call action on Elasticsearch document indexer"""
        socket = self.indexer.get_socket()
        if socket is None:
            return [501, "No socket handler defined in configuration file"]
        if action in ('index_document', 'update_document'):
            intids = get_utility(IIntIds)
            document = intids.queryId(document)
        settings = {
            'zodb_name': self.indexer.zodb_name,
            'document': document
        }
        settings.update(kwargs)
        LOGGER.debug(f"Sending JSON message: action={action}, {settings}")
        socket.send_json([action, settings])
        return zmq_response(socket)


@factory_config(IContentIndexerUtility)
class ContentIndexerUtility(Persistent, Contained):
    """Elasticsearch content index utility"""

    zodb_name = FieldProperty(IContentIndexerUtility['zodb_name'])

    @staticmethod
    def get_socket():
        """Open ØMQ socket"""
        registry = get_pyramid_registry()
        handler = registry.settings.get(INDEXER_HANDLER_KEY, False)
        if handler:
            return zmq_socket(handler, auth=registry.settings.get(INDEXER_AUTH_KEY))
        return None

    def test_process(self):
        """Send test request to document indexer process"""
        socket = self.get_socket()
        if socket is None:
            return [501, "No socket handler defined in configuration file"]
        socket.send_json(['test', {}])
        return zmq_response(socket)

    def index_document(self, document):
        """Add or update document to Elasticsearch index"""
        handler = IndexUpdateHandler(self)
        return handler.update_index('index_document', document)

    def update_document(self, document, attrs):
        """Update document attributes in Elasticsearch index"""
        handler = IndexUpdateHandler(self)
        return handler.update_index('update_document', document, attrs=attrs)

    def unindex_document(self, document_oid):
        """Remove document from Elasticsearch index"""
        handler = IndexUpdateHandler(self)
        return handler.update_index('unindex_document', document_oid)


#
# Backoffice quick search settings
#

@factory_config(IQuickSearchSettings)
class QuickSearchSettings(Persistent, Contained):
    """Quick search settings persistent class"""

    analyzer = FieldProperty(IQuickSearchSettings['analyzer'])
    search_fields = FieldProperty(IQuickSearchSettings['search_fields'])
    default_operator = FieldProperty(IQuickSearchSettings['default_operator'])


@adapter_config(required=IContentIndexerUtility,
                provides=IQuickSearchSettings)
def content_indexer_quick_search_settings(context):
    """Content indexer quick search settings"""
    return get_annotation_adapter(context, QUICK_SEARCH_SETTINGS_KEY,
                                  IQuickSearchSettings)


#
# Default user search settings
#

@factory_config(IUserSearchSettings)
class UserSearchSettings(Persistent, Contained):
    """User search settings persistent class"""

    analyzer = FieldProperty(IUserSearchSettings['analyzer'])
    search_fields = FieldProperty(IUserSearchSettings['search_fields'])
    fulltext_search_fields = FieldProperty(IUserSearchSettings['fulltext_search_fields'])
    default_operator = FieldProperty(IUserSearchSettings['default_operator'])


@adapter_config(required=IContentIndexerUtility,
                provides=IUserSearchSettings)
def content_indexer_user_search_settings(context):
    """Content indexer user search settings"""
    return get_annotation_adapter(context, USER_SEARCH_SETTINGS_KEY,
                                  IUserSearchSettings)



#
# Copyright (c) 2015-2024 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

from pyams_content.component.frame.interfaces import IFrameParagraph
from pyams_content_es.component import get_index_values, html_to_index
from pyams_content_es.component.paragraph import base_paragraph_index_info
from pyams_content_es.interfaces import IDocumentIndexInfo
from pyams_utils.adapter import adapter_config

__docformat__ = 'restructuredtext'


@adapter_config(required=IFrameParagraph,
                provides=IDocumentIndexInfo)
def frame_paragraph_index_info(context):
    """Frame paragraph index info"""
    info = base_paragraph_index_info(context)
    get_index_values(context, info,
                     i18n_fields=(('body', html_to_index),))
    return info



#
# Copyright (c) 2015-2024 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

from pyams_content.feature.filter import ICollectionsFilter, IContentTypesFilter, IFilterAggregate, ITagsFilter, \
    IThemesFilter, ITitleFilter
from pyams_content.feature.filter.interfaces import FILTER_SORTING
from pyams_i18n.interfaces import INegotiator
from pyams_thesaurus.interfaces.thesaurus import IThesaurus, IThesaurusExtracts
from pyams_utils.adapter import adapter_config
from pyams_utils.registry import get_pyramid_registry, get_utility, query_utility

__docformat__ = 'restructuredtext'


SORTING_PARAMS = {
    FILTER_SORTING.ALPHA.value:      {'_key': 'asc'},
    FILTER_SORTING.ALPHA_DESC.value: {'_key': 'desc'},
    FILTER_SORTING.COUNT.value:      {'_count': 'asc'},
    FILTER_SORTING.COUNT_DESC.value: {'_count': 'desc'}
}


def get_sorting_params(sorting_mode):
    """Returns the sorting parameters for Elasticsearch queries based on the provided sorting option.

    :param sorting_mode: A string representing the desired sorting option.
    :return: A dictionary representing the Elasticsearch sorting parameters.
    """
    return SORTING_PARAMS.get(sorting_mode, {'_key': 'asc'})


@adapter_config(required=IContentTypesFilter,
                provides=IFilterAggregate)
def content_type_filter_aggregate(context):
    """Content-type filter aggregate getter"""
    content_mode = context.content_mode
    sorting_params = get_sorting_params(context.sorting_mode)
    registry = get_pyramid_registry()
    return {
        'name': content_mode,
        'type': 'terms',
        'params': {
            'field': registry.settings.get(f'pyams_content_es.filter.{content_mode}.field_name',
                                           content_mode),
            'size': 100,
            'order': sorting_params
        }
    }


@adapter_config(required=ITitleFilter,
                provides=IFilterAggregate)
def title_filter_aggregate(context):
    """Title filter aggregate getter"""
    sorting_params = get_sorting_params(context.sorting_mode)
    registry = get_pyramid_registry()
    field_name = registry.settings.get('pyams_content_es.filter.title.field_name')
    if not field_name:
        negotiator = get_utility(INegotiator)
        field_name = f'title.{negotiator.server_language}.keyword'
    return {
        'name': 'title',
        'type': 'terms',
        'params': {
            'field': field_name,
            'size': 100,
            'order': sorting_params
        }
    }


def get_terms_aggregate(context, field_name):
    """Helper function used to build thesaurus filter aggregate"""
    thesaurus = query_utility(IThesaurus, name=context.thesaurus_name)
    if thesaurus is None:
        return None
    sorting_params = get_sorting_params(context.sorting_mode)
    registry = get_pyramid_registry()
    filter_aggregate = {
        'name': context.filter_type,
        'type': 'terms',
        'params': {
            'field': registry.settings.get(f'pyams_content_es.filter.{field_name}.field_name',
                                           field_name),
            'size': 100,
            'order': sorting_params
        }
    }
    if context.extract_name:
        extract = IThesaurusExtracts(thesaurus).get(context.extract_name)
        if extract is not None:
            terms = list(extract.terms_ids)
            if terms:
                filter_aggregate['params']['include'] = terms
    return filter_aggregate


@adapter_config(required=ITagsFilter,
                provides=IFilterAggregate)
def tags_filter_aggregate(context):
    """tags filter aggregate getter"""
    return get_terms_aggregate(context, 'tags')


@adapter_config(required=IThemesFilter,
                provides=IFilterAggregate)
def theme_filter_aggregate(context):
    """theme filter aggregate getter"""
    return get_terms_aggregate(context, 'themes.terms')


@adapter_config(required=ICollectionsFilter,
                provides=IFilterAggregate)
def collection_filter_aggregate(context):
    """Content-type filter aggregate getter"""
    return get_terms_aggregate(context, 'collections')



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.zmi.content module

This module provides components which are used to check a specific content against
Elasticsearch index.
"""

__docformat__ = 'restructuredtext'

from dateutil import parser
from elasticsearch_dsl import Search
from zope.interface import Interface, implementer
from zope.schema import TextLine
from zope.schema.fieldproperty import FieldProperty

from pyams_content.feature.search import ISearchFolder
from pyams_content.interfaces import MANAGE_CONTENT_PERMISSION
from pyams_content_es import _
from pyams_content_es.interfaces import IDocumentIndexTarget
from pyams_form.ajax import ajax_form_config
from pyams_form.button import Buttons, handler
from pyams_form.field import Fields
from pyams_form.interfaces.form import IAJAXFormRenderer, IFormContent
from pyams_i18n.interfaces import II18n
from pyams_layer.interfaces import IPyAMSLayer
from pyams_skin.interfaces.viewlet import IFormHeaderViewletManager
from pyams_skin.schema.button import CloseButton, SubmitButton
from pyams_skin.viewlet.help import AlertMessage
from pyams_skin.viewlet.menu import MenuItem
from pyams_utils.adapter import ContextRequestViewAdapter, NullAdapter, adapter_config
from pyams_utils.date import format_datetime
from pyams_utils.timezone import tztime
from pyams_viewlet.viewlet import viewlet_config
from pyams_workflow.interfaces import IWorkflow
from pyams_zmi.form import AdminModalEditForm
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.viewlet import IContextActionsDropdownMenu


@viewlet_config(name='content-index.menu',
                context=IDocumentIndexTarget, layer=IAdminLayer,
                manager=IContextActionsDropdownMenu, weight=30,
                permission=MANAGE_CONTENT_PERMISSION)
class ContentIndexCheckerMenu(MenuItem):
    """Content index checker menu"""

    label = _("Check index content")
    icon_class = 'fas fa-check'

    href = 'content-index.html'
    modal_target = True


@viewlet_config(name='content-index.menu',
                context=ISearchFolder, layer=IAdminLayer,
                manager=IContextActionsDropdownMenu, weight=30,
                permission=MANAGE_CONTENT_PERMISSION)
class SearchFolderIndexCheckerMenu(NullAdapter):
    """Search folder index checker menu"""


class IContentIndexCheckerFormFields(Interface):
    """Content index checker form fields interface"""

    id = TextLine(title=_("Internal ID"),
                  readonly=True)

    timestamp = TextLine(title=_("Modification date"),
                         readonly=True)

    status = TextLine(title=_("Workflow status"),
                      readonly=True)


@implementer(IContentIndexCheckerFormFields)
class ContentIndexCheckerFormContent:
    """Content index checker form content"""

    id = FieldProperty(IContentIndexCheckerFormFields['id'])
    timestamp = FieldProperty(IContentIndexCheckerFormFields['timestamp'])
    status = FieldProperty(IContentIndexCheckerFormFields['status'])


class IContentIndexCheckerFormButtons(Interface):
    """Content index checker form buttons interface"""

    reindex = SubmitButton(name='reindex',
                           title=_("Force reindex"))

    close = CloseButton(name='close',
                        title=_("Cancel"))


@ajax_form_config(name='content-index.html',
                  context=IDocumentIndexTarget, layer=IPyAMSLayer,
                  permission=MANAGE_CONTENT_PERMISSION)
class ContentIndexCheckerForm(AdminModalEditForm):
    """Content index checker form"""

    subtitle = _("Index content check")
    legend = _("Current index content")

    fields = Fields(IContentIndexCheckerFormFields)
    buttons = Buttons(IContentIndexCheckerFormButtons)

    _edit_permission = MANAGE_CONTENT_PERMISSION

    @handler(buttons['reindex'])
    def handle_reindex(self, action):
        """Reindex button handle"""
        self.handle_apply(self, action)

    def apply_changes(self, data):
        client = self.request.elastic_client
        client.index_object(self.context)


@viewlet_config(name='help',
                context=IDocumentIndexTarget, layer=IAdminLayer,
                view=ContentIndexCheckerForm,
                manager=IFormHeaderViewletManager, weight=10)
class ContentIndexCheckerFormHelp(AlertMessage):
    """Content index checker form help"""

    status = 'info'

    _message = _("Documents informations are stored in an Elasticsearch index. If an error "
                 "occurs, some documents can be stored in this index with different values from "
                 "those stored into contents database.<br />"
                 "If required, this operation will try to reindex document properties into "
                 "Elasticsearch.")
    message_renderer = 'markdown'


@adapter_config(required=(IDocumentIndexTarget, IAdminLayer, ContentIndexCheckerForm),
                provides=IFormContent)
def document_index_checker_form_content(context, request, form):
    """Document index checker form content getter"""
    client = request.elastic_client
    query = Search(using=client.es, index=client.index) \
        .query('term', _id=context.id) \
        .source(['_id', '@timestamp', 'workflow'])
    content = ContentIndexCheckerFormContent()
    translate = request.localizer.translate
    for result in query.execute().hits:
        content.id = result.meta.id
        content.timestamp = format_datetime(tztime(parser.parse(result['@timestamp'])))
        content.status = translate(IWorkflow(context).get_state_label(result.workflow.status))
    return content


@adapter_config(required=(IDocumentIndexTarget, IAdminLayer, ContentIndexCheckerForm),
                provides=IAJAXFormRenderer)
class ContentIndexCheckerFormRenderer(ContextRequestViewAdapter):
    """Content index checker form renderer"""

    def render(self, changes):
        translate = self.request.localizer.translate
        return {
            'status': 'success',
            'message': translate(_("Request successful, index content has been updated"))
        }



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.process module

This module defines PyAMS_content indexing process.
"""
from multiprocessing import Process

from pprint import pformat
from pyramid.threadlocal import RequestContext
from threading import Thread
from transaction.interfaces import ITransactionManager
from zope.interface import implementer
from zope.intid import IIntIds

from pyams_content_es.interfaces import IContentIndexerProcess, LOGGER
from pyams_elastic.include import client_from_config
from pyams_site.interfaces import PYAMS_APPLICATION_DEFAULT_NAME, PYAMS_APPLICATION_SETTINGS_KEY
from pyams_utils.registry import get_pyramid_registry, get_utility, set_local_registry
from pyams_utils.request import check_request
from pyams_utils.transaction import COMMITTED_STATUS
from pyams_utils.zodb import ZODBConnection
from pyams_zmq.handler import ZMQMessageHandler
from pyams_zmq.process import ZMQProcess


__docformat__ = 'restructuredtext'


class BaseIndexUpdaterProcess(Process):
    """Base Elasticsearch index updater process"""

    def __init__(self, registry, settings, group=None, target=None, name=None, *args, **kwargs):
        super().__init__(group, target, name, *args, **kwargs)
        self.registry = registry
        self.settings = settings

    def run(self):
        """Index update process"""
        LOGGER.debug("Starting Elastincsearch index updater process...")
        # check settings
        settings = self.settings
        LOGGER.debug(f"Checking parameters: {settings}")
        document_id = settings.pop('document', None)
        if not document_id:
            LOGGER.warning("Elasticsearch index updater: missing document ID!")
            return
        # loading components registry
        registry = self.registry
        LOGGER.debug(f"Getting Pyramid registry: {registry!r}")
        # get ES client
        es_client = client_from_config(registry.settings)
        if es_client is None:
            LOGGER.warning("Missing Elasticsearch client configuration!")
            return
        # create new request
        request = check_request()
        request.registry = registry
        LOGGER.debug(f"Creating new request {request!r}")
        with RequestContext(request):
            # open ZODB connection
            LOGGER.debug("Opening ZODB connection...")
            zodb_name = settings.pop('zodb_name', '')
            with ZODBConnection(zodb_name) as root:
                LOGGER.debug(f"Getting database root {root!r}")
                application_name = registry.settings.get(PYAMS_APPLICATION_SETTINGS_KEY,
                                                         PYAMS_APPLICATION_DEFAULT_NAME)
                application = root.get(application_name)
                LOGGER.debug(f"Loaded application {application!r} "
                             f"with name {application_name}")
                if application is not None:
                    # set local registry
                    sm = application.getSiteManager()
                    set_local_registry(sm)
                    LOGGER.debug(f"Setting local registry {sm!r}")
                    self.update_index(es_client, document_id, **settings)

    def update_index(self, client, document_id, **kwargs):
        """Update Elasticsearch index"""
        raise NotImplementedError("Index updater process must implement update_index method!")


class DocumentIndexProcess(BaseIndexUpdaterProcess):
    """Document indexer process"""

    def update_index(self, client, document_id, **kwargs):
        """Add document to Elasticsearch index

        Document ID must be provided as an internal reference ID, as provided
        by IntIds utility.
        """
        # search document
        intids = get_utility(IIntIds)
        document = intids.queryObject(document_id)
        if document is None:
            LOGGER.warning(f"Can't find requested document {document_id}!")
            return
        # index document
        LOGGER.debug(f"Starting indexing for {document!r}")
        manager = ITransactionManager(document, None)
        try:
            for attempt in manager.attempts():
                with attempt as t:
                    client.index_object(document)
                if t.status == COMMITTED_STATUS:
                    break
        finally:
            if manager is not None:
                manager.abort()


class DocumentUpdaterProcess(BaseIndexUpdaterProcess):
    """Document updater process"""

    def update_index(self, client, document_id, **kwargs):
        """Update document attributes in Elasticsearch index

        Document ID must be provided as an Elasticsearch ID, as document as
        already been removed from ZODB.
        """
        intids = get_utility(IIntIds)
        document = intids.queryObject(document_id)
        if document is None:
            LOGGER.warning(f"Can't find requested document {document_id}!")
            return
        # update document
        attrs = kwargs.pop('attrs', ())
        LOGGER.debug(f"Starting updating for {document!r}")
        manager = ITransactionManager(document, None)
        try:
            for attempt in manager.attempts():
                with attempt as t:
                    client.update_object(document, attrs, **kwargs)
                if t.status == COMMITTED_STATUS:
                    break
        finally:
            if manager is not None:
                manager.abort()


class DocumentUnindexProcess(BaseIndexUpdaterProcess):
    """Document un-index process"""

    def update_index(self, client, document_id, **kwargs):
        """Remove document from Elasticsearch index

        Document ID must be provided as an Elasticsearch ID, as document as
        already been removed from ZODB.
        """
        LOGGER.debug(f"Starting removal for document {document_id}")
        intids = get_utility(IIntIds)
        manager = ITransactionManager(intids, None)
        try:
            for attempt in manager.attempts():
                with attempt as t:
                    client.delete_document(document_id, safe=True)
                if t.status == COMMITTED_STATUS:
                    break
        finally:
            if manager is not None:
                manager.abort()


class ContentIndexerThread(Thread):
    """Content indexer thread"""

    def __init__(self, process):
        super().__init__()
        self.process = process

    def run(self):
        self.process.start()
        self.process.join()


class ContentIndexerHandler:
    """Elasticsearch indexer handler"""

    process = None

    @staticmethod
    def ping(settings):
        """Content indexer ping handler"""
        return [200, 'pong']

    @staticmethod
    def test(settings):  # pylint: disable=unused-argument
        """Content indexer test handler"""
        messages = [
            'OK - Documents indexer process ready to handle requests.', ''
        ]
        registry = get_pyramid_registry()
        es_client = client_from_config(registry.settings)
        if es_client is None:
            messages.append('WARNING: no Elasticsearch client available!')
        else:
            messages.extend([
                "Elasticsearch client properties:",
                " - nodes:"
            ])
            for config in es_client.es.transport.node_pool.node_selector.node_configs:
                messages.extend([
                    f"{' ' * 6}{line}"
                    for line in pformat(config, indent=5).split('\n')
                ])
            messages.extend([
                f" - index name: {es_client.index}",
                f" - indexing: {'DISABLED' if es_client.disable_indexing else 'enabled'}",
                ""
            ])
            ping = es_client.es.ping()
            messages.append(f"Server ping: {'OK' if ping else 'KO'}")
            if ping:
                messages.extend(['', 'Server info:'])
                messages.extend(pformat(es_client.es.info().body).split('\n'))
        return [200, '\n'.join(messages)]

    @staticmethod
    def index_document(settings):
        """Add or update document in Elasticsearch index"""
        registry = get_pyramid_registry()
        ContentIndexerThread(DocumentIndexProcess(registry, settings)).start()
        return [200, 'Content indexer process started']

    @staticmethod
    def update_document(settings):
        """Update document attributes in Elasticsearch index"""
        registry = get_pyramid_registry()
        ContentIndexerThread(DocumentUpdaterProcess(registry, settings)).start()

    @staticmethod
    def unindex_document(settings):
        """Remove document from Elasticsearch index"""
        registry = get_pyramid_registry()
        ContentIndexerThread(DocumentUnindexProcess(registry, settings)).start()
        return [200, 'Content un-index process started']


class ContentIndexerMessageHandler(ZMQMessageHandler):
    """Elasticsearch indexer message handler"""

    handler = ContentIndexerHandler


@implementer(IContentIndexerProcess)
class ContentIndexerProcess(ZMQProcess):
    """Elasticsearch indexer process"""



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

__docformat__ = 'restructuredtext'

from elasticsearch_dsl import Q
from zope.intid import IIntIds
from zope.schema.vocabulary import getVocabularyRegistry

from pyams_content.root.zmi.dashboard import SiteRootArchivedContentsTable, \
    SiteRootArchivedContentsValues, SiteRootDashboardManagerWaitingTable, \
    SiteRootDashboardManagerWaitingValues, SiteRootDashboardOwnerModifiedTable, \
    SiteRootDashboardOwnerModifiedValues, SiteRootDashboardOwnerWaitingTable, \
    SiteRootDashboardOwnerWaitingValues, SiteRootLastModificationsTable, \
    SiteRootLastModificationsValues, SiteRootLastPublicationsTable, \
    SiteRootLastPublicationsValues, SiteRootPreparationsTable, SiteRootPreparationsValues, \
    SiteRootPublicationsTable, SiteRootPublicationsValues, SiteRootRetiredContentsTable, \
    SiteRootRetiredContentsValues, SiteRootSubmissionsTable, SiteRootSubmissionsValues
from pyams_content.shared.common.interfaces import IBaseSharedTool, SHARED_CONTENT_TYPES_VOCABULARY
from pyams_content.shared.common.zmi.dashboard import SharedToolArchivedContentsTable, \
    SharedToolArchivedContentsValues, SharedToolDashboardManagerWaitingTable, \
    SharedToolDashboardManagerWaitingValues, SharedToolDashboardOwnerWaitingTable, \
    SharedToolDashboardOwnerWaitingValues, SharedToolLastModificationsTable, \
    SharedToolLastModificationsValues, SharedToolLastPublicationsTable, \
    SharedToolLastPublicationsValues, SharedToolPreparationsTable, SharedToolPreparationsValues, \
    SharedToolPublicationsTable, SharedToolPublicationsValues, SharedToolRetiredContentsTable, \
    SharedToolRetiredContentsValues, SharedToolSubmissionsTable, SharedToolSubmissionsValues
from pyams_content_es.document import ElasticResultSet, get_elastic_search
from pyams_site.interfaces import ISiteRoot
from pyams_table.interfaces import IValues
from pyams_utils.adapter import adapter_config
from pyams_utils.list import unique_iter
from pyams_utils.registry import get_all_utilities_registered_for, get_utility
from pyams_workflow.interfaces import IWorkflow
from pyams_workflow.versions import get_last_version_in_state
from pyams_zmi.interfaces import IAdminLayer


MODIFIED_DATE_FIELD = 'workflow.modified_date'


#
# Elasticsearch shared tools dashboards adapters
#

@adapter_config(required=(IBaseSharedTool, IAdminLayer, SharedToolDashboardManagerWaitingTable),
                provides=IValues)
class EsSharedToolDashboardManagerWaitingValues(SharedToolDashboardManagerWaitingValues):
    """Elasticsearch shared tool dashboard manager waiting values getter"""

    @property
    def values(self):
        """Table values adapter"""
        intids = get_utility(IIntIds)
        workflow = IWorkflow(self.context)
        vocabulary = getVocabularyRegistry().get(self.context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = (
            Q('term', parent_ids=intids.register(self.context)) &
            Q('terms', content_type=list(vocabulary.by_value.keys())) &
            Q('terms', workflow__status=list(workflow.waiting_states)))
        search = get_elastic_search(self.request, params)
        yield from filter(
            self.check_access,
            unique_iter(map(get_last_version_in_state, ElasticResultSet(search))))


@adapter_config(required=(IBaseSharedTool, IAdminLayer, SharedToolDashboardOwnerWaitingTable),
                provides=IValues)
class EsSharedToolDashboardOwnerWaitingValues(SharedToolDashboardOwnerWaitingValues):
    """Elasticsearch shared tool dashboard waiting owned contents values adapter"""

    @property
    def values(self):
        """Table values getter"""
        principal_id = self.request.principal.id
        intids = get_utility(IIntIds)
        workflow = IWorkflow(self.context)
        vocabulary = getVocabularyRegistry().get(self.context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = (
            Q('term', parent_ids=intids.register(self.context)) &
            Q('terms', content_type=list(vocabulary.by_value.keys())) &
            Q('terms', workflow__status=list(workflow.waiting_states)) &
            Q('term', workflow__principal=principal_id))
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(map(get_last_version_in_state, ElasticResultSet(search)))


@adapter_config(required=(IBaseSharedTool, IAdminLayer, SharedToolPreparationsTable),
                provides=IValues)
class EsSharedToolPreparationsValues(SharedToolPreparationsValues):
    """Elasticsearch shared tool preparations values adapter"""

    @property
    def values(self):
        principal_id = self.request.principal.id
        intids = get_utility(IIntIds)
        workflow = IWorkflow(self.context)
        vocabulary = getVocabularyRegistry().get(self.context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = (
            Q('term', parent_ids=intids.register(self.context)) &
            Q('terms', content_type=list(vocabulary.by_value.keys())) &
            (Q('term', owner_id=principal_id) |
             Q('term', contributor_id=principal_id)) &
            Q('term', workflow__status=workflow.initial_state)
        )
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search))


@adapter_config(required=(IBaseSharedTool, IAdminLayer, SharedToolSubmissionsTable),
                provides=IValues)
class EsSharedToolSubmissionsValues(SharedToolSubmissionsValues):
    """Elasticsearch shared tool submissions values adapter"""

    @property
    def values(self):
        context = self.context
        principal_id = self.request.principal.id
        intids = get_utility(IIntIds)
        workflow = IWorkflow(context)
        vocabulary = getVocabularyRegistry().get(context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = (
            Q('term', parent_ids=intids.register(self.context)) &
            Q('terms', content_type=list(vocabulary.by_value.keys())) &
            (Q('term', owner_id=principal_id) |
             Q('term', contributor_id=principal_id)) &
            Q('terms', workflow__status=list(workflow.waiting_states))
        )
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search))


@adapter_config(required=(IBaseSharedTool, IAdminLayer, SharedToolPublicationsTable),
                provides=IValues)
class EsSharedToolPublicationsValues(SharedToolPublicationsValues):
    """Elasticsearch shared tool publications values adapter"""

    @property
    def values(self):
        context = self.context
        principal_id = self.request.principal.id
        intids = get_utility(IIntIds)
        workflow = get_utility(IWorkflow, name=context.shared_content_workflow)
        vocabulary = getVocabularyRegistry().get(context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = (
            Q('term', parent_ids=intids.register(self.context)) &
            Q('terms', content_type=list(vocabulary.by_value.keys())) &
            (Q('term', owner_id=principal_id) |
             Q('term', contributor_id=principal_id)) &
            Q('terms', workflow__status=list(workflow.published_states))
        )
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search))


@adapter_config(required=(IBaseSharedTool, IAdminLayer, SharedToolRetiredContentsTable),
                provides=IValues)
class EsSharedToolRetiredContentsValues(SharedToolRetiredContentsValues):
    """Elasticsearch shared tool retired contents values adapter"""

    @property
    def values(self):
        context = self.context
        principal_id = self.request.principal.id
        intids = get_utility(IIntIds)
        workflow = get_utility(IWorkflow, name=context.shared_content_workflow)
        vocabulary = getVocabularyRegistry().get(context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = (
            Q('term', parent_ids=intids.register(self.context)) &
            Q('terms', content_type=list(vocabulary.by_value.keys())) &
            (Q('term', owner_id=principal_id) |
             Q('term', contributor_id=principal_id)) &
            Q('terms', workflow__status=list(workflow.retired_states))
        )
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search))


@adapter_config(required=(IBaseSharedTool, IAdminLayer, SharedToolArchivedContentsTable),
                provides=IValues)
class EsSharedToolArchivedContentsValues(SharedToolArchivedContentsValues):
    """Elasticsearch shared tool archived contents values adapter"""

    @property
    def values(self):
        context = self.context
        principal_id = self.request.principal.id
        intids = get_utility(IIntIds)
        workflow = get_utility(IWorkflow, name=context.shared_content_workflow)
        vocabulary = getVocabularyRegistry().get(context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = (
            Q('term', parent_ids=intids.register(self.context)) &
            Q('terms', content_type=list(vocabulary.by_value.keys())) &
            (Q('term', owner_id=principal_id) |
             Q('term', contributor_id=principal_id)) &
            Q('terms', workflow__status=list(workflow.archived_states))
        )
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search))


@adapter_config(required=(IBaseSharedTool, IAdminLayer, SharedToolLastPublicationsTable),
                provides=IValues)
class EsSharedToolLastPublicationsValues(SharedToolLastPublicationsValues):
    """Elasticsearch shared tool publications values adapter"""

    @property
    def values(self):
        intids = get_utility(IIntIds)
        workflow = get_utility(IWorkflow, name=self.context.shared_content_workflow)
        vocabulary = getVocabularyRegistry().get(self.context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = (
            Q('term', parent_ids=intids.register(self.context)) &
            Q('terms', content_type=list(vocabulary.by_value.keys())) &
            Q('terms', workflow__status=list(workflow.published_states))
        )
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search, 50))


@adapter_config(required=(IBaseSharedTool, IAdminLayer, SharedToolLastModificationsTable),
                provides=IValues)
class EsSharedToolLastModificationsValues(SharedToolLastModificationsValues):
    """Elasticsearch shared tool modifications values adapter"""

    @property
    def values(self):
        intids = get_utility(IIntIds)
        vocabulary = getVocabularyRegistry().get(self.context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = (
            Q('term', parent_ids=intids.register(self.context)) &
            Q('terms', content_type=list(vocabulary.by_value.keys()))
        )
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search, 50))


#
# Elasticsearch site root dashboards adapters
#

@adapter_config(required=(ISiteRoot, IAdminLayer, SiteRootDashboardManagerWaitingTable),
                provides=IValues)
class EsSiteRootDashboardManagerWaitingValues(SiteRootDashboardManagerWaitingValues):
    """Elasticsearch site root dashboard waiting values adapter"""

    @property
    def values(self):
        """Table values getter"""
        intids = get_utility(IIntIds)
        vocabulary = getVocabularyRegistry().get(self.context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = None
        for tool in get_all_utilities_registered_for(IBaseSharedTool):
            workflow = IWorkflow(tool)
            query = (
                Q('term', parent_ids=intids.register(tool)) &
                Q('terms', content_type=list(vocabulary.by_value.keys())) &
                Q('terms', workflow__status=list(workflow.waiting_states))
            )
            params = params | query if params else query
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from filter(
            self.check_access,
            unique_iter(map(get_last_version_in_state, ElasticResultSet(search))))


@adapter_config(required=(ISiteRoot, IAdminLayer, SiteRootDashboardOwnerWaitingTable),
                provides=IValues)
class EsSiteRootDashboardOwnerWaitingValues(SiteRootDashboardOwnerWaitingValues):
    """Elasticsearch site root dashboard waiting owned contents values adapter"""

    @property
    def values(self):
        """Table values getter"""
        principal_id = self.request.principal.id
        intids = get_utility(IIntIds)
        vocabulary = getVocabularyRegistry().get(self.context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = None
        for tool in get_all_utilities_registered_for(IBaseSharedTool):
            workflow = IWorkflow(tool)
            query = (
                Q('term', parent_ids=intids.register(tool)) &
                Q('terms', content_type=list(vocabulary.by_value.keys())) &
                Q('terms', workflow__status=list(workflow.waiting_states)) &
                Q('term', workflow__principal=principal_id)
            )
            params = params | query if params else query
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(map(get_last_version_in_state, ElasticResultSet(search)))


@adapter_config(required=(ISiteRoot, IAdminLayer, SiteRootDashboardOwnerModifiedTable),
                provides=IValues)
class EsSiteRootDashboardOwnerModifiedValues(SiteRootDashboardOwnerModifiedValues):
    """Elasticsearch site root dashboard owner modified adapter"""

    @property
    def values(self):
        """Table values getter"""
        principal_id = self.request.principal.id
        intids = get_utility(IIntIds)
        vocabulary = getVocabularyRegistry().get(self.context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = None
        for tool in get_all_utilities_registered_for(IBaseSharedTool):
            query = (
                Q('term', parent_ids=intids.register(tool)) &
                Q('terms', content_type=list(vocabulary.by_value.keys())) &
                (Q('term', owner_id=principal_id) |
                 Q('term', contributor_id=principal_id))
            )
            params = params | query if params else query
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(map(get_last_version_in_state, ElasticResultSet(search, 50)))


@adapter_config(required=(ISiteRoot, IAdminLayer, SiteRootPreparationsTable),
                provides=IValues)
class EsSiteRootPreparationsValues(SiteRootPreparationsValues):
    """Elasticsearch site root preparations values adapter"""

    @property
    def values(self):
        principal_id = self.request.principal.id
        intids = get_utility(IIntIds)
        vocabulary = getVocabularyRegistry().get(self.context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = None
        for tool in get_all_utilities_registered_for(IBaseSharedTool):
            workflow = IWorkflow(tool)
            query = (
                Q('term', parent_ids=intids.register(tool)) &
                Q('terms', content_type=list(vocabulary.by_value.keys())) &
                (Q('term', owner_id=principal_id) |
                 Q('term', contributor_id=principal_id)) &
                Q('term', workflow__status=workflow.initial_state)
            )
            params = params | query if params else query
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search))


@adapter_config(required=(ISiteRoot, IAdminLayer, SiteRootSubmissionsTable),
                provides=IValues)
class EsSiteRootSubmissionsValues(SiteRootSubmissionsValues):
    """Elasticsearch site root submissions values adapter"""

    @property
    def values(self):
        context = self.context
        principal_id = self.request.principal.id
        intids  =get_utility(IIntIds)
        vocabulary = getVocabularyRegistry().get(context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = None
        for tool in get_all_utilities_registered_for(IBaseSharedTool):
            workflow = IWorkflow(tool)
            query = (
                Q('term', parent_ids=intids.register(tool)) &
                Q('terms', content_type=list(vocabulary.by_value.keys())) &
                (Q('term', owner_id=principal_id) |
                 Q('term', contributor_id=principal_id)) &
                Q('terms', workflow__status=list(workflow.waiting_states))
            )
            params = params | query if params else query
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search))


@adapter_config(required=(ISiteRoot, IAdminLayer, SiteRootPublicationsTable),
                provides=IValues)
class EsSiteRootPublicationsValues(SiteRootPublicationsValues):
    """Elasticsearch site root publications values adapter"""

    @property
    def values(self):
        context = self.context
        principal_id = self.request.principal.id
        intids = get_utility(IIntIds)
        vocabulary = getVocabularyRegistry().get(context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = None
        for tool in get_all_utilities_registered_for(IBaseSharedTool):
            workflow = IWorkflow(tool)
            query = (
                Q('term', parent_ids=intids.register(tool)) &
                Q('terms', content_type=list(vocabulary.by_value.keys())) &
                (Q('term', owner_id=principal_id) |
                 Q('term', contributor_id=principal_id)) &
                Q('terms', workflow__status=list(workflow.published_states))
            )
            params = params | query if params else query
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search))


@adapter_config(required=(ISiteRoot, IAdminLayer, SiteRootRetiredContentsTable),
                provides=IValues)
class EsSiteRootRetiredContentsValues(SiteRootRetiredContentsValues):
    """Elasticsearch site root retired contents values adapter"""

    @property
    def values(self):
        context = self.context
        principal_id = self.request.principal.id
        intids = get_utility(IIntIds)
        vocabulary = getVocabularyRegistry().get(context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = None
        for tool in get_all_utilities_registered_for(IBaseSharedTool):
            workflow = IWorkflow(tool)
            query = (
                Q('term', parent_ids=intids.register(tool)) &
                Q('terms', content_type=list(vocabulary.by_value.keys())) &
                (Q('term', owner_id=principal_id) |
                 Q('term', contributor_id=principal_id)) &
                Q('terms', workflow__status=list(workflow.retired_states))
            )
            params = params | query if params else query
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search))


@adapter_config(required=(ISiteRoot, IAdminLayer, SiteRootArchivedContentsTable),
                provides=IValues)
class EsSiteRootArchivedContentsValues(SiteRootArchivedContentsValues):
    """Elasticsearch site root archived contents values adapter"""

    @property
    def values(self):
        context = self.context
        principal_id = self.request.principal.id
        intids = get_utility(IIntIds)
        vocabulary = getVocabularyRegistry().get(context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = None
        for tool in get_all_utilities_registered_for(IBaseSharedTool):
            workflow = IWorkflow(tool)
            query = (
                Q('term', parent_ids=intids.register(tool)) &
                Q('terms', content_type=list(vocabulary.by_value.keys())) &
                (Q('term', owner_id=principal_id) |
                 Q('term', contributor_id=principal_id)) &
                Q('terms', workflow__status=list(workflow.archived_states))
            )
            params = params | query if params else query
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search))


@adapter_config(required=(ISiteRoot, IAdminLayer, SiteRootLastPublicationsTable),
                provides=IValues)
class EsSiteRootLastPublicationsValues(SiteRootLastPublicationsValues):
    """Elasticsearch site root publications values adapter"""

    @property
    def values(self):
        intids = get_utility(IIntIds)
        vocabulary = getVocabularyRegistry().get(self.context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = None
        for tool in get_all_utilities_registered_for(IBaseSharedTool):
            workflow = IWorkflow(tool)
            query = (
                Q('term', parent_ids=intids.register(tool)) &
                Q('terms', content_type=list(vocabulary.by_value.keys())) &
                Q('terms', workflow__status=list(workflow.published_states))
            )
            params = params | query if params else query
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search, 50))


@adapter_config(required=(ISiteRoot, IAdminLayer, SiteRootLastModificationsTable),
                provides=IValues)
class EsSiteRootLastModificationsValues(SiteRootLastModificationsValues):
    """Elasticsearch site root modifications values adapter"""

    @property
    def values(self):
        vocabulary = getVocabularyRegistry().get(self.context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = Q('terms', content_type=list(vocabulary.by_value.keys()))
        search = get_elastic_search(self.request, params) \
            .sort({
                MODIFIED_DATE_FIELD: {
                    'unmapped_type': 'date',
                    'order': 'desc'
                }
            })
        yield from unique_iter(ElasticResultSet(search, 50))



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.zmi.search module

This module is used to provide components which are used by search engines.
"""

from elasticsearch_dsl import Q, Search
from zope.interface import Interface
from zope.intid import IIntIds
from zope.schema import Bool, TextLine
from zope.schema.vocabulary import getVocabularyRegistry

from pyams_content.root.zmi.search import ISiteRootAdvancedSearchQuery, \
    SiteRootAdvancedSearchForm, SiteRootAdvancedSearchResultsTable, \
    SiteRootAdvancedSearchResultsValues, SiteRootQuickSearchResultsTable, \
    SiteRootQuickSearchResultsValues
from pyams_content.shared.common import IBaseSharedTool, SHARED_CONTENT_TYPES_VOCABULARY
from pyams_content.shared.common.zmi.search import ISharedToolAdvancedSearchQuery, \
    SharedToolAdvancedSearchForm, SharedToolAdvancedSearchResultsTable, \
    SharedToolAdvancedSearchResultsValues, SharedToolQuickSearchResultsTable, \
    SharedToolQuickSearchResultsValues
from pyams_content_es.document import ElasticResultSet
from pyams_content_es.interfaces import IContentIndexerUtility, IQuickSearchSettings, IUserSearchSettings
from pyams_elastic.include import get_client
from pyams_form.field import Fields
from pyams_form.interfaces.form import IFormFields
from pyams_layer.interfaces import IPyAMSLayer
from pyams_sequence.interfaces import ISequentialIntIds
from pyams_sequence.workflow import get_last_version
from pyams_site.interfaces import ISiteRoot
from pyams_table.interfaces import IValues
from pyams_utils.adapter import adapter_config
from pyams_utils.list import unique_iter
from pyams_utils.registry import get_utility
from pyams_workflow.versions import get_last_version_in_state
from pyams_zmi.interfaces import IAdminLayer

__docformat__ = 'restructuredtext'

from pyams_content_es import _


def get_search_params(data):
    """Search params getter"""
    intids = get_utility(IIntIds)
    query = data.get('query')
    if query:
        sequence = get_utility(ISequentialIntIds)
        if query.startswith('+'):
            yield Q('term',
                    reference_id=sequence.get_full_oid(query))
        else:
            indexer = get_utility(IContentIndexerUtility)
            settings = IUserSearchSettings(indexer)
            fulltext = data.get('fulltext', False)
            if fulltext:
                yield (
                    Q('term',
                      reference_id=sequence.get_full_oid(query)) |
                    Q('simple_query_string',
                      query=query,
                      fields=settings.fulltext_search_fields,
                      analyzer=settings.analyzer,
                      default_operator=settings.default_operator,
                      lenient=True))
            else:
                yield (
                    Q('term',
                      reference_id=sequence.get_full_oid(query)) |
                    Q('simple_query_string',
                      query=query,
                      fields=settings.search_fields,
                      analyzer=settings.analyzer,
                      default_operator=settings.default_operator))
    if data.get('owner'):
        yield Q('term',
                owner_id=data['owner'])
    if data.get('status'):
        yield Q('term',
                workflow__status=data['status'])
    if data.get('content_type'):
        yield Q('term',
                content_type=data['content_type'])
    if data.get('data_type'):
        yield Q('term',
                data_type=data['data_type'])
    created_after, created_before = data.get('created', (None, None))
    if created_after:
        yield Q('range',
                workflow__created_date={'gte': created_after})
    if created_before:
        yield Q('range',
                workflow__created_date={'lte': created_before})
    modified_after, modified_before = data.get('modified', (None, None))
    if modified_after:
        yield Q('range',
                workflow__modified_date={'gte': modified_after})
    if modified_before:
        yield Q('range',
                workflow__modified_date={'lte': modified_before})
    if data.get('tags'):
        tags = [intids.register(term) for term in data['tags']]
        yield Q('terms',
                tags=tags)
    if data.get('themes'):
        themes = [intids.register(term) for term in data['themes']]
        yield Q('terms',
                themes__terms=themes)
    if data.get('collections'):
        collections = [intids.register(collection) for collection in data['collections']]
        yield Q('terms',
                collections=collections)


#
# Custom shared tools quick search adapters
#

@adapter_config(required=(IBaseSharedTool, IPyAMSLayer, SharedToolQuickSearchResultsTable),
                provides=IValues)
class EsSharedToolQuickSearchResultsValues(SharedToolQuickSearchResultsValues):
    """Elasticsearch shared tool quick search results values adapter"""

    @property
    def values(self):
        """Table values getter"""
        intids = get_utility(IIntIds)
        query = self.request.params.get('query', '').strip()
        if not query:
            return ()
        sequence = get_utility(ISequentialIntIds)
        query = query.lower()
        if query.startswith('+'):
            params = Q('term',
                       reference_id=sequence.get_full_oid(query))
        else:
            indexer = get_utility(IContentIndexerUtility)
            settings = IQuickSearchSettings(indexer)
            vocabulary = getVocabularyRegistry().get(self.context,
                                                     SHARED_CONTENT_TYPES_VOCABULARY)
            params = (
                Q('term',
                  parent_ids=intids.register(self.context)) &
                Q('terms',
                  content_type=list(vocabulary.by_value.keys())) &
                (Q('term',
                   reference_id=sequence.get_full_oid(query)) |
                 Q('simple_query_string',
                   query=query,
                   fields=settings.search_fields,
                   analyzer=settings.analyzer,
                   default_operator=settings.default_operator)))
        client = get_client(self.request)
        search = Search(using=client.es, index=client.index) \
            .query(params) \
            .source(['internal_id'])
        yield from unique_iter(map(get_last_version, ElasticResultSet(search)))


#
# Custom shared tools advanced search adapters
#

class IEsSharedToolAdvancedSearchQuery(ISharedToolAdvancedSearchQuery):
    """Elasticsearch shared tool advanced search form fields interface"""

    query = TextLine(title=_("Search text"),
                     description=_("Entered text will be search in titles, headers and "
                                   "descriptions"),
                     required=False)

    fulltext = Bool(title=_("Fulltext search"),
                    description=_("Search in fulltext body, including attachments, instead on "
                                  "only searching in titles and headers"))


@adapter_config(required=(Interface, IAdminLayer, SharedToolAdvancedSearchForm),
                provides=IFormFields)
def es_shared_tool_advanced_search_form_fields(context, request, form):
    """Elasticsearch shared tool advanced search form fields getter"""
    return Fields(IEsSharedToolAdvancedSearchQuery).select('query', 'fulltext') + \
        Fields(IEsSharedToolAdvancedSearchQuery).omit('query', 'fulltext',
                                                      'tags', 'themes', 'collections')


@adapter_config(required=(IBaseSharedTool, IPyAMSLayer, SharedToolAdvancedSearchResultsTable),
                provides=IValues)
class EsSharedToolAdvancedSearchResultsValues(SharedToolAdvancedSearchResultsValues):
    """Elasticsearch shared tool advanced search results adapter"""

    def get_params(self, data):
        """Extract Elasticsearch query params from incoming request"""
        intids = get_utility(IIntIds)
        vocabulary = getVocabularyRegistry().get(self.context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = Q('term',
                   parent_ids=intids.register(self.context)) & \
            Q('terms',
              content_type=list(vocabulary.by_value.keys()))
        for param in get_search_params(data):
            params &= param
        return params

    @property
    def values(self):
        """Query values getter"""
        form = SharedToolAdvancedSearchForm(self.context, self.request)
        form.update()
        data, _errors = form.extract_data()
        params = self.get_params(data)
        client = get_client(self.request)
        search = Search(using=client.es, index=client.index) \
            .query(params) \
            .source(['internal_id'])
        sort_order = [{
            'workflow.modified_date': {
                'unmapped_type': 'date',
                'order': 'desc'
            }
        }]
        if data.get('query'):
            sort_order.insert(0, {
                '_score': {
                    'order': 'desc'
                }
            })
        search = search.sort(*sort_order)
        if data.get('status'):
            yield from unique_iter(map(get_last_version_in_state, ElasticResultSet(search)))
        else:
            yield from unique_iter(map(get_last_version, ElasticResultSet(search)))


#
# Custom site root quick search results adapters
#

@adapter_config(required=(ISiteRoot, IPyAMSLayer, SiteRootQuickSearchResultsTable),
                provides=IValues)
class EsSiteRootQuickSearchResultsValues(SiteRootQuickSearchResultsValues):
    """Elasticsearch site root quick search results table values adapter"""

    @property
    def values(self):
        """Table values getter"""
        query = self.request.params.get('query', '').strip()
        if not query:
            return ()
        sequence = get_utility(ISequentialIntIds)
        query = query.lower()
        if query.startswith('+'):
            params = Q('term',
                       reference_id=sequence.get_full_oid(query))
        else:
            vocabulary = getVocabularyRegistry().get(self.context,
                                                     SHARED_CONTENT_TYPES_VOCABULARY)
            indexer = get_utility(IContentIndexerUtility)
            settings = IQuickSearchSettings(indexer)
            params = (
                Q('terms',
                  content_type=list(vocabulary.by_value.keys())) &
                (Q('term',
                   reference_id=sequence.get_full_oid(query)) |
                 Q('simple_query_string',
                   query=query,
                   fields=settings.search_fields,
                   analyzer=settings.analyzer,
                   default_operator=settings.default_operator)))
        client = get_client(self.request)
        search = Search(using=client.es, index=client.index) \
            .query(params) \
            .source(['internal_id'])
        yield from unique_iter(map(get_last_version, ElasticResultSet(search)))


#
# Custom site root advanced search adapters
#

class IEsSiteRootAdvancedSearchQuery(ISiteRootAdvancedSearchQuery):
    """Elasticsearch site root advanced search form fields interfaces"""

    query = TextLine(title=_("Search text"),
                     description=_("Entered text will be search in titles, headers and "
                                   "descriptions"),
                     required=False)

    fulltext = Bool(title=_("Fulltext search"),
                    description=_("Search in fulltext body, including attachments, instead on "
                                  "only searching in titles and headers"))


@adapter_config(required=(ISiteRoot, IAdminLayer, SiteRootAdvancedSearchForm),
                provides=IFormFields)
def es_site_root_advanced_search_form_fields(context, request, form):
    """Elasticsearch site root advanced search form fields getter"""
    return Fields(IEsSiteRootAdvancedSearchQuery).select('query', 'fulltext') + \
        Fields(IEsSiteRootAdvancedSearchQuery).omit('query', 'fulltext',
                                                    'tags', 'themes', 'collections')


@adapter_config(required=(ISiteRoot, IPyAMSLayer, SiteRootAdvancedSearchResultsTable),
                provides=IValues)
class EsSiteRootAdvancedSearchResultsValues(SiteRootAdvancedSearchResultsValues):
    """Elasticsearch site root advanced search results adapter"""

    def get_params(self, data):
        """Extract Elasticsearch query params from incoming request"""
        vocabulary = getVocabularyRegistry().get(self.context, SHARED_CONTENT_TYPES_VOCABULARY)
        params = Q('terms',
                   content_type=list(vocabulary.by_value.keys()))
        for param in get_search_params(data):
            params &= param
        return params

    @property
    def values(self):
        form = SiteRootAdvancedSearchForm(self.context, self.request)
        form.update()
        data, _errors = form.extract_data()
        params = self.get_params(data)
        client = get_client(self.request)
        search = Search(using=client.es, index=client.index) \
            .query(params) \
            .source(['internal_id'])
        sort_order = [{
            'workflow.modified_date': {
                'unmapped_type': 'date',
                'order': 'desc'
            }
        }]
        if data.get('query'):
            sort_order.insert(0, {
                '_score': {
                    'order': 'desc'
                }
            })
        search = search.sort(*sort_order)
        yield from unique_iter(map(get_last_version, ElasticResultSet(search)))



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.interfaces module

"""

import logging
from zope.annotation.interfaces import IAttributeAnnotatable
from zope.interface import Interface
from zope.schema import Choice, TextLine

from pyams_utils.interfaces import ZODB_CONNECTIONS_VOCABULARY_NAME
from pyams_utils.schema import TextLineListField
from pyams_zmq.interfaces import IZMQProcess

from pyams_content_es import _


#
# Content indexer utility
#

LOGGER = logging.getLogger('PyAMS (ES content)')


INDEXER_NAME = 'Elasticsearch indexer'
INDEXER_LABEL = _("Elasticsearch documents indexer")


INDEXER_PREFIX = 'pyams_content_es'
INDEXER_STARTER_KEY = f'{INDEXER_PREFIX}.start_handler'
INDEXER_HANDLER_KEY = f'{INDEXER_PREFIX}.tcp_handler'
INDEXER_AUTH_KEY = f'{INDEXER_PREFIX}.allow_auth'
INDEXER_CLIENTS_KEY = f'{INDEXER_PREFIX}.allow_clients'


class IContentIndexerProcess(IZMQProcess):
    """Content indexer process"""


class IContentIndexerHandler(Interface):
    """Content indexer handler interface"""


class IContentIndexerUtility(IAttributeAnnotatable):
    """Content indexer utility interface"""

    zodb_name = Choice(title=_("ZODB connection name"),
                       description=_("Name of ZODB defining document indexer connection"),
                       required=False,
                       default='',
                       vocabulary=ZODB_CONNECTIONS_VOCABULARY_NAME)

    def get_socket(self):
        """ØMQ socket getter"""

    def test_process(self):
        """Test document indexer process connection"""

    def index_document(self, document):
        """Add or update document to Elasticsearch index"""

    def unindex_document(self, document):
        """Remove document from Elasticsearch index"""


class IDocumentIndexInfo(Interface):
    """Document index info"""


class IDocumentIndexTarget(Interface):
    """Document index target marker interface"""


#
# Search settings interfaces
#

class IBaseSearchSettings(Interface):
    """Base search settings interface"""

    analyzer = TextLine(title=_("Search analyzer"),
                        description=_("Used search analyzer; check index settings to get list of "
                                      "available analyzers..."),
                        required=True,
                        default='default')

    default_operator = Choice(title=_("Default operator"),
                              description=_("This is the default operator used to combine search terms"),
                              values=('AND', 'OR'),
                              required=True,
                              default='AND')


QUICK_SEARCH_SETTINGS_KEY = 'pyams_content_es.quick_settings'


class IQuickSearchSettings(IBaseSearchSettings):
    """Backoffice quick search settings interface"""

    search_fields = TextLineListField(title=_("Search fields"),
                                      description=_("List of fields used for quick search in backoffice, "
                                                    "with their respective weight (if any)"),
                                      required=True,
                                      default=[
                                          'title.*',
                                          'short_name.*',
                                          'header.*',
                                          'description.*'
                                      ])


USER_SEARCH_SETTINGS_KEY = 'pyams_content_es.user_settings'


class IUserSearchSettings(IBaseSearchSettings):
    """Default user search settings"""

    search_fields = TextLineListField(title=_("Backoffice search fields"),
                                      description=_("List of fields used for base search in backoffice, "
                                                    "with their respective weight (if any)"),
                                      required=True,
                                      default=[
                                          'title.*',
                                          'short_name.*',
                                          'header.*',
                                          'description.*'
                                      ])

    fulltext_search_fields = TextLineListField(title=_("Fulltext search fields"),
                                               description=_("List of fields used for fulltext search, "
                                                             "with their respective weight (if any)"),
                                               required=True,
                                               default=[
                                                   'title.*',
                                                   'short_name.*',
                                                   'header.*',
                                                   'description.*'
                                               ])



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.component.workflow module

This module defines adapters which are used to handle indexation of
workflow status information.
"""

__docformat__ = 'restructuredtext'

from pyramid.events import subscriber
from transaction.interfaces import ITransactionManager
from zope.dublincore.interfaces import IZopeDublinCore
from zope.intid import IIntIds

from pyams_content_es.document import INDEXED_DOCUMENTS_EXTENSION
from pyams_content_es.interfaces import IContentIndexerUtility, IDocumentIndexInfo, \
    IDocumentIndexTarget
from pyams_utils.adapter import adapter_config
from pyams_utils.registry import get_utility, query_utility
from pyams_workflow.interfaces import IWorkflowInfo, IWorkflowPublicationInfo, \
    IWorkflowPublicationSupport, IWorkflowState, IWorkflowTransitionEvent


@adapter_config(name='workflow',
                required=IWorkflowPublicationSupport,
                provides=IDocumentIndexInfo)
def workflow_managed_index_info(content):
    """Workflow managed content index info"""
    result = {}
    dc_info = IZopeDublinCore(content, None)
    if dc_info is not None:
        result.update({
            'created_date': dc_info.created,
            'modified_date': dc_info.modified
        })
    workflow_info = IWorkflowInfo(content, None)
    if workflow_info is not None:
        result.update({
            'name': workflow_info.name
        })
    workflow_pub_info = IWorkflowPublicationInfo(content, None)
    if workflow_pub_info is not None:
        result.update({
            'publication_date': workflow_pub_info.publication_date,
            'effective_date': workflow_pub_info.publication_effective_date,
            'push_end_date': workflow_pub_info.push_end_date,
            'expiration_date': workflow_pub_info.publication_expiration_date,
            'first_publication_date': workflow_pub_info.first_publication_date,
            'content_publication_date': workflow_pub_info.content_publication_date,
            'visible_publication_date': workflow_pub_info.visible_publication_date
        })
    workflow_state = IWorkflowState(content, None)
    if workflow_state is not None:
        result.update({
            'status': workflow_state.state,
            'principal': workflow_state.state_principal,
            'date': workflow_state.state_date,
        })
    return {
        'workflow': result
    }


@subscriber(IWorkflowTransitionEvent, context_selector=IDocumentIndexTarget)
def handle_workflow_transition(event):
    """Handle workflow transition

    When a workflow transition occurs, just update document workflow information
    instead of the whole document.
    """
    indexer = query_utility(IContentIndexerUtility)
    if indexer is None:
        return
    intids = get_utility(IIntIds)
    document = event.object
    document_id = intids.register(document)
    transaction = ITransactionManager(document).get()
    documents = transaction.extension.get(INDEXED_DOCUMENTS_EXTENSION) or set()
    if document_id not in documents:
        documents.add(document_id)
        transaction.extension[INDEXED_DOCUMENTS_EXTENSION] = documents
        indexer.update_document(document, ['workflow'])



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.zmi module

This module defines a few components used to handle Elasticsearch content
indexer utility properties.
"""

from zope.interface import Interface

from pyams_content_es.interfaces import IContentIndexerUtility, INDEXER_LABEL, IQuickSearchSettings, IUserSearchSettings
from pyams_form.ajax import ajax_form_config
from pyams_form.button import Buttons, handler
from pyams_form.field import Fields
from pyams_form.interfaces.form import IFormContent, IGroup, IInnerSubForm
from pyams_form.subform import InnerDisplayForm
from pyams_layer.interfaces import IPyAMSLayer
from pyams_security.interfaces.base import MANAGE_SYSTEM_PERMISSION
from pyams_skin.interfaces.view import IModalPage
from pyams_skin.schema.button import ActionButton
from pyams_template.template import template_config
from pyams_utils.adapter import adapter_config
from pyams_utils.url import absolute_url
from pyams_zmi.form import AdminModalDisplayForm, AdminModalEditForm, FormGroupSwitcher
from pyams_zmi.interfaces import IAdminLayer, IObjectLabel
from pyams_zmi.interfaces.form import IFormTitle, IModalEditFormButtons
from pyams_zmi.interfaces.table import ITableElementEditor
from pyams_zmi.table import TableElementEditor
from pyams_zmi.utils import get_object_label

__docformat__ = 'restructuredtext'

from pyams_content_es import _


@adapter_config(required=(IContentIndexerUtility, IPyAMSLayer, Interface),
                provides=IObjectLabel)
def content_indexer_label(context, request, view):
    """Content indexer label"""
    return request.localizer.translate(INDEXER_LABEL)


@adapter_config(required=(IContentIndexerUtility, IAdminLayer, Interface),
                provides=ITableElementEditor)
class ContentIndexerElementEditor(TableElementEditor):
    """Content index element editor"""

    def __new__(cls, context, request, view):
        if not request.has_permission(MANAGE_SYSTEM_PERMISSION, context=context):
            return None
        return TableElementEditor.__new__(cls)


class IContentIndexerPropertiesEditFormButtons(IModalEditFormButtons):
    """Content indexer properties edit form buttons"""

    test = ActionButton(name='test',
                        title=_("Test connection"))


@ajax_form_config(name='properties.html',
                  context=IContentIndexerUtility, layer=IPyAMSLayer,
                  permission=MANAGE_SYSTEM_PERMISSION)
class ContentIndexerPropertiesEditForm(AdminModalEditForm):
    """Content indexer properties edit form"""

    modal_class = 'modal-xl'
    legend = _("Document indexer properties")
    fields = Fields(IContentIndexerUtility)
    buttons = Buttons(IContentIndexerPropertiesEditFormButtons).select('test', 'apply', 'close')

    def update_actions(self):
        super().update_actions()
        test = self.actions.get('test')
        if test is not None:
            test.add_class('btn-info mr-auto')
            test.href = absolute_url(self.context, self.request, 'test-indexer.html')
            test.modal_target = True

    @handler(IContentIndexerPropertiesEditFormButtons['apply'])
    def handle_apply(self, action):
        """Apply button handler"""
        super().handle_apply(self, action)


@adapter_config(required=(IContentIndexerUtility, IAdminLayer, IModalPage),
                provides=IFormTitle)
def content_indexer_form_title(context, request, view):
    return get_object_label(context, request, view)


class ContentIndexerSettingsGroup(FormGroupSwitcher):
    """Content indexer settings group"""

    switcher_mode = 'always'

    def update_widgets(self, prefix=None, use_form_mode=True):
        super().update_widgets(prefix, use_form_mode)
        fields = self.widgets.get('search_fields')
        if fields is not None:
            fields.rows = 8


@adapter_config(name='quick-settings',
                required=(IContentIndexerUtility, IAdminLayer, ContentIndexerPropertiesEditForm),
                provides=IGroup)
class ContentIndexerQuickSearchSettingsGroup(ContentIndexerSettingsGroup):
    """Content index quick search settings group"""

    legend = _("Quick search settings")

    prefix = 'quick_settings.'
    fields = Fields(IQuickSearchSettings)
    weight = 10


@adapter_config(required=(IContentIndexerUtility, IAdminLayer, ContentIndexerQuickSearchSettingsGroup),
                provides=IFormContent)
def content_indexer_quick_search_settings_content(context, request, group):
    """Content indexer quick search settings edit form group content getter"""
    return IQuickSearchSettings(context)


@adapter_config(name='user-settings',
                required=(IContentIndexerUtility, IAdminLayer, ContentIndexerPropertiesEditForm),
                provides=IGroup)
class ContentIndexerUserSearchSettingsGroup(ContentIndexerSettingsGroup):
    """Content index user search settings group"""

    legend = _("User search settings")

    prefix = 'user_settings.'
    fields = Fields(IUserSearchSettings).select('analyzer', 'search_fields',
                                                'fulltext_search_fields', 'default_operator')
    weight = 20

    def update_widgets(self, prefix=None, use_form_mode=True):
        super().update_widgets(prefix, use_form_mode)
        fields = self.widgets.get('fulltext_search_fields')
        if fields is not None:
            fields.rows = 8


@adapter_config(required=(IContentIndexerUtility, IAdminLayer, ContentIndexerUserSearchSettingsGroup),
                provides=IFormContent)
def content_indexer_user_search_group_content(context, request, group):
    """Content indexer user search settings edit form group content getter"""
    return IUserSearchSettings(context)


#
# Indexer test form
#

@ajax_form_config(name='test-indexer.html',
                  context=IContentIndexerUtility, layer=IPyAMSLayer,
                  permission=MANAGE_SYSTEM_PERMISSION)
class ContentIndexerTestForm(AdminModalDisplayForm):
    """Content indexer test form"""


@adapter_config(name='test',
                required=(IContentIndexerUtility, IAdminLayer, ContentIndexerTestForm),
                provides=IInnerSubForm)
@template_config(template='templates/test-indexer.pt', layer=IAdminLayer)
class ContentIndexerTestStatusForm(InnerDisplayForm):
    """Content indexer test status form"""

    legend = _("Content indexer test")

    def test(self):
        """Content indexer test"""
        return self.context.test_process()



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.component.thesaurus module

This module defines adapters which are used to handle thesaurus terms indexation.
"""

__docformat__ = 'restructuredtext'

from zope.intid import IIntIds

from pyams_content.component.thesaurus import ICollectionsInfo, ICollectionsTarget, ITagsInfo, \
    ITagsTarget, IThemesInfo, IThemesTarget
from pyams_content_es.interfaces import IDocumentIndexInfo
from pyams_utils.adapter import adapter_config
from pyams_utils.list import unique
from pyams_utils.registry import get_utility


def get_term_ids(terms):
    """Term ID getter"""
    intids = get_utility(IIntIds)
    return list(map(intids.queryId, unique(terms)))


@adapter_config(name='tags',
                required=ITagsTarget,
                provides=IDocumentIndexInfo)
def tags_index_info(context):
    """Tags document index info"""
    tags = ITagsInfo(context).tags or ()
    return {
        'tags': get_term_ids(tags)
    }


@adapter_config(name='themes',
                required=IThemesTarget,
                provides=IDocumentIndexInfo)
def themes_index_info(context):
    """Themes document index info"""
    terms= []
    parents = []
    synonyms = []
    associations = []
    for term in IThemesInfo(context).themes or ():
        terms.append(term)
        if term.usage is not None:
            terms.append(term.usage)
        parents.extend(term.get_parents())
        synonyms.extend(term.used_for)
        associations.extend(term.associations)
    return {
        'themes': {
            'terms': get_term_ids(terms),
            'parents': get_term_ids(parents),
            'synonyms': get_term_ids(synonyms),
            'associations': get_term_ids(associations)
        }
    }


@adapter_config(name='collections',
                required=ICollectionsTarget,
                provides=IDocumentIndexInfo)
def collections_index_info(context):
    """Collections document index info"""
    collections = ICollectionsInfo(context).collections or ()
    return {
        'collections': get_term_ids(collections)
    }



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.component main module

"""

from pyams_i18n.interfaces import II18nManager, INegotiator
from pyams_utils.html import html_to_text
from pyams_utils.registry import get_utility

__docformat__ = 'restructuredtext'


_marker = object()


def html_to_index(value):
    """HTML converter to index value"""
    return html_to_text(value).replace(chr(13), '')


def get_index_values(context, output, fields=None, i18n_fields=None):
    """Get index values from context"""
    i18n_manager = II18nManager(context, None)
    if i18n_manager is None:
        negotiator = get_utility(INegotiator)
        languages = negotiator.offered_languages
    else:
        languages = i18n_manager.get_languages()
    # simple fields getter
    for field in (fields or ()):
        if isinstance(field, tuple):
            field, converter = field
        else:
            converter = None
        value = getattr(context, field, _marker)
        if (not value) or (value is _marker):
            continue
        if converter is not None:
            value = converter(value)
        for lang in languages:
            output[lang] = f"{output.get(lang, '')}\n" \
                           f"{value}"
    # i18n fields getter
    for field in (i18n_fields or ()):
        if isinstance(field, tuple):
            field, converter = field
        else:
            converter = None
        value = getattr(context, field, _marker)
        if (not value) or (value is _marker):
            continue
        for lang, i18n_value in value.items():
            if not i18n_value:
                continue
            if converter is not None:
                i18n_value = converter(i18n_value)
            output[lang] = f"{output.get(lang, '')}\n" \
                           f"{i18n_value}"



#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.shared.view module

This module defines how shared content queries are handled when using an Elasticsearch
index.
"""

__docformat__ = 'restructuredtext'

from elasticsearch_dsl import A, AttrDict, Q, Search
from zope.interface import implementer

from pyams_catalog.query import CatalogResultSet
from pyams_content.shared.view import IViewQuery, IWfView
from pyams_content.shared.view.interfaces import RELEVANCE_ORDER, TITLE_ORDER
from pyams_content.shared.view.interfaces.query import END_PARAMS_MARKER
from pyams_content_es.interfaces import IContentIndexerUtility, IUserSearchSettings
from pyams_content_es.shared.view.interfaces import IEsViewQuery, IEsViewQueryParamsExtension, \
    IEsViewQueryPostFilterExtension, IEsViewQueryResultsFilterExtension, IEsViewUserPostFilter, IEsViewUserQuery
from pyams_elastic.include import get_client
from pyams_i18n.interfaces import INegotiator
from pyams_sequence.interfaces import ISequentialIntIds
from pyams_utils.adapter import ContextAdapter, adapter_config, get_adapter_weight
from pyams_utils.list import boolean_iter, unique_iter
from pyams_utils.registry import get_pyramid_registry, get_utility
from pyams_workflow.interfaces import IWorkflow


@adapter_config(required=IWfView,
                provides=IViewQuery)
@implementer(IEsViewQuery)
class EsViewQuery(ContextAdapter):
    """Elasticsearch view query"""

    def get_params(self, context, request=None, **kwargs):
        view = self.context
        registry = get_pyramid_registry()
        # check publication dates
        params = Q('range',
                   **{'workflow.effective_date': {'lte': 'now/m'}})
        # check workflow states
        if 'state' in kwargs:
            state = kwargs['state']
            if not isinstance(state, (list, tuple, set)):
                state = (state,)
            params &= Q('terms',
                        **{'workflow.status': state})
        else:
            wf_params = []
            for workflow in registry.getAllUtilitiesRegisteredFor(IWorkflow):
                wf_params.extend(workflow.visible_states)
            params &= Q('terms',
                        **{'workflow.status': wf_params})
        # check custom extensions
        get_all_params = True
        for name, adapter in sorted(registry.getAdapters((view,), IEsViewQueryParamsExtension),
                                    key=get_adapter_weight):
            for new_params in adapter.get_params(context, request):
                if new_params is None:
                    return None
                elif new_params is END_PARAMS_MARKER:
                    get_all_params = False
                    break
                else:
                    params &= new_params
            else:
                continue
            break
        # activate search
        params &= Q('bool',
                    must=Q('range',
                           **{'workflow.push_end_date': {'gte': 'now/m'}})) | \
                  Q('bool',
                    must_not=Q('exists',
                               **{'field': 'workflow.push_end_date'}))
        if get_all_params:
            # check content path
            content_path = view.get_content_path(context)
            if content_path is not None:
                params &= Q('term',
                            **{'parent_ids': content_path})
            # check content types
            if 'content_type' in kwargs:
                params &= Q('term',
                            **{'content_type': kwargs['content_type']})
            else:
                params &= Q('bool',
                            must_not=Q('terms',
                                       **{'content_type': tuple(view.get_ignored_types())}))
                content_types = view.get_content_types(context)
                if content_types:
                    params &= Q('terms',
                                **{'content_type': content_types})
            # check data types
            data_types = view.get_data_types(context)
            if data_types:
                params &= Q('terms',
                            **{'data_type': data_types})
            # check excluded content types
            content_types = view.get_excluded_content_types(context)
            if content_types:
                params &= Q('bool',
                            must_not=Q('terms',
                                       **{'content_type': content_types}))
            # check excluded data types
            data_types = view.get_excluded_data_types(context)
            if data_types:
                params &= Q('bool',
                            must_not=Q('terms',
                                       **{'data_type': data_types}))
            # check age limit
            age_limit = view.age_limit
            if age_limit:
                params &= Q('range',
                            **{
                                'workflow.content_publication_date': {
                                    'gte': f'now-{age_limit}d/m'
                                }
                            })
        return params

    def get_post_filters(self, context, request=None, **kwargs):
        registry = get_pyramid_registry()
        for name, adapter in sorted(registry.getAdapters((self.context,), IEsViewQueryPostFilterExtension),
                                    key=get_adapter_weight):
            yield from adapter.get_params(context, request)

    def get_results(self, context, sort_index, reverse, limit,
                    request=None, aggregates=None, settings=None, **kwargs):
        aggregations = {}
        registry = request.registry
        client = get_client(request)
        params = self.get_params(context, request, **kwargs)
        if params is None:
            items = CatalogResultSet([])
            total_count = 0
        else:
            # get post-filters (excluded from aggregations calculation)
            filters = Q()
            for filter in self.get_post_filters(context, request, **kwargs):
                filters &= filter
            search = Search(using=client.es, index=client.index) \
                .params(request_timeout=30) \
                .query(params) \
                .post_filter(filters) \
                .source(['internal_id'])
            if aggregates:
                for agg in aggregates:
                    search.aggs.bucket(agg['name'],
                                       A(agg['type'], **agg['params']))
            # Define sort order
            sort_values = []
            if (not sort_index) or (sort_index == RELEVANCE_ORDER):
                sort_values.append({
                    '_score': {
                        'order': 'desc'
                    }
                })
            elif sort_index == TITLE_ORDER:
                negotiator = get_utility(INegotiator)
                sort_values.append({
                    f'title.{negotiator.server_language}.keyword': {
                        'order': 'desc' if reverse else 'asc'
                    }
                })
            else:
                sort_values.append({
                    f'workflow.{sort_index}': {
                        'order': 'desc' if reverse else 'asc',
                        'unmapped_type': 'date'
                    }
                })
            if sort_values:
                search = search.sort(*sort_values)
            # Define limits
            if limit:
                search = search[:limit]
            else:
                search = search[:999]
            # Get query results
            results = search.execute()
            items = CatalogResultSet([result.internal_id for result in results.hits])
            aggregations = results.aggregations
            total_count = results.hits.total
            if isinstance(total_count, (dict, AttrDict)):
                total_count = results.hits.total['value']
        for name, adapter in sorted(registry.getAdapters((self.context,),
                                                         IEsViewQueryResultsFilterExtension),
                                    key=lambda x: x[1].weight):
            items = adapter.filter(context, items, request)
        return total_count, aggregations, unique_iter(items)


class BaseEsUserViewQueryExtension(ContextAdapter):
    """Base Elasticsearch user view query extension"""

    weight = 999
    query_interface = None

    def __new__(cls, context):
        if not context.allow_user_params:
            return None
        return ContextAdapter.__new__(cls)

    def get_params(self, context, request=None):
        """User params getter"""
        registry = get_pyramid_registry()
        for name, adapter in sorted(registry.getAdapters((self.context,), self.query_interface),
                                    key=get_adapter_weight):
            yield from adapter.get_user_params(request)


@adapter_config(name='user-params',
                required=IWfView,
                provides=IEsViewQueryParamsExtension)
class EsUserViewQueryParamsExtension(BaseEsUserViewQueryExtension):
    """Elasticsearch user view query params extension"""

    query_interface = IEsViewUserQuery


@adapter_config(name='user-params',
                required=IWfView,
                provides=IEsViewQueryPostFilterExtension)
class EsUserViewQueryPostFilterExtension(BaseEsUserViewQueryExtension):
    """Elasticsearch user view query post-filters extension"""

    query_interface = IEsViewUserPostFilter


class EsViewSimpleTermQuery(ContextAdapter):
    """Elasticsearch simple view term query adapter"""

    param_name = None
    field_name = None

    def get_user_params(self, request):
        params = request.params.getall(self.param_name)
        if params:
            registry = request.registry
            field_name = registry.settings.get(f'pyams_content_es.filter.{self.field_name}.field_name',
                                               self.field_name)
            yield Q('terms', **{field_name: params})


@adapter_config(name='content-type',
                required=IWfView,
                provides=IEsViewUserQuery)
class EsViewContentTypeQuery(EsViewSimpleTermQuery):
    """Search folder content-type query"""

    param_name = 'content_type'
    field_name = 'content_type'


@adapter_config(name='data-type',
                required=IWfView,
                provides=IEsViewUserQuery)
class EsViewDatatypeQuery(EsViewSimpleTermQuery):
    """Search folder data-type query"""

    param_name = 'data_type'
    field_name = 'data_type'


@adapter_config(name='facet-label',
                required=IWfView,
                provides=IEsViewUserQuery)
class EsViewFacetLabelQuery(EsViewSimpleTermQuery):
    """Search folder facet label query"""

    param_name = 'facet_label'
    field_name = 'facet_label'


@adapter_config(name='facet-type-label',
                required=IWfView,
                provides=IEsViewUserQuery)
class EsViewFacetTypeLabelQuery(EsViewSimpleTermQuery):
    """Search folder facet type label query"""

    param_name = 'facet_type_label'
    field_name = 'facet_type_label'


@adapter_config(name='title',
                required=IWfView,
                provides=IEsViewUserPostFilter)
class EsViewTitleQuery(EsViewSimpleTermQuery):
    """Search folder title query"""

    param_name = 'title'

    @property
    def field_name(self):
        negotiator = get_utility(INegotiator)
        return f'title.{negotiator.server_language}.keyword'


@adapter_config(name='user-search',
                required=IWfView,
                provides=IEsViewUserQuery)
class EsViewUserSearchQuery(ContextAdapter):
    """Elasticsearch user search query adapter"""

    @staticmethod
    def get_user_params(request):
        """User params getter"""
        fulltext = request.params.get('user_search')
        if fulltext:
            if fulltext.startswith('+'):
                sequence = get_utility(ISequentialIntIds)
                oid = sequence.get_full_oid(fulltext)
                yield Q('term',
                        **{'reference_id': oid})
            else:
                indexer = get_utility(IContentIndexerUtility)
                settings = IUserSearchSettings(indexer)
                query = {
                    'query': fulltext,
                    'default_operator': settings.default_operator,
                    'analyzer': settings.analyzer,
                    'fields': settings.fulltext_search_fields,
                    'lenient': True
                }
                if '"' not in fulltext:
                    query2 = query.copy()
                    query2['query'] = '"{}"'.format(fulltext)
                    yield (Q('simple_query_string', **query) |
                           Q('simple_query_string', **query2))
                else:
                    yield Q('simple_query_string', **query)



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

__docformat__ = 'restructuredtext'

from datetime import datetime, timezone
from elasticsearch_dsl import Search

from pyramid.events import subscriber
from transaction.interfaces import ITransactionManager
from zope.dublincore.interfaces import IZopeDublinCore
from zope.interface import classImplements
from zope.intid import IIntIds
from zope.lifecycleevent import ObjectModifiedEvent
from zope.lifecycleevent.interfaces import IObjectAddedEvent, IObjectModifiedEvent, IObjectRemovedEvent

from pyams_catalog.query import CatalogResultSet
from pyams_content.shared.common import WfSharedContent
from pyams_content.shared.common.interfaces import IPreventSharedContentUpdateSubscribers, ISharedTool, \
    IWfSharedContent, IWfSharedContentRoles
from pyams_content.shared.common.interfaces.types import IWfTypedSharedContent
from pyams_content_es.interfaces import IContentIndexerUtility, IDocumentIndexInfo, IDocumentIndexTarget
from pyams_elastic.include import get_client
from pyams_elastic.mixin import ESField, ESKeyword, ESMapping, ESText, ElasticMixin
from pyams_i18n.interfaces import II18n
from pyams_sequence.interfaces import ISequentialIdInfo
from pyams_utils.adapter import adapter_config
from pyams_utils.interfaces.traversing import IPathElements
from pyams_utils.registry import get_pyramid_registry, get_utility, query_utility
from pyams_utils.request import query_request
from pyams_utils.traversing import get_parent
from pyams_workflow.interfaces import IWorkflowState


def get_internal_id(doc):
    """Internal ID getter"""
    return doc.internal_id


def get_elastic_search(request, params, index=None):
    """Create Elasticsearch query from given params"""
    client = get_client(request)
    return Search(using=client.es,
                  index=index or client.index) \
        .query(params) \
        .source(['internal_id'])


def ElasticResultSet(search, length=999):
    """Elasticsearch documents results set

    Returns an iterator on documents from an Elasticsearch query
    containing documents internal IDs.
    """
    return CatalogResultSet(map(get_internal_id, search[:length]))


class ElasticDocumentMixin(ElasticMixin):
    """Elasticsearch document mixin class"""

    @property
    def id(self):
        """Document ID getter"""
        oid = ISequentialIdInfo(self).hex_oid
        state = IWorkflowState(self, None)
        if state is None:
            return oid
        return f'{oid}.{state.version_id}'

    @property
    def timestamp(self):
        """Timestamp getter"""
        dc = IZopeDublinCore(self, None)
        if dc is None:
            return datetime.now(timezone.utc)
        return dc.modified

    @property
    def internal_id(self):
        """Internal ID getter"""
        intids = get_utility(IIntIds)
        return intids.register(self)

    @property
    def reference_id(self):
        """Reference ID getter"""
        return ISequentialIdInfo(self).hex_oid

    @property
    def owner_id(self):
        """Owner ID getter"""
        return list(IWfSharedContentRoles(self).owner or ())

    @property
    def contributor_id(self):
        """Contributor IDs getter"""
        return list(IWfSharedContentRoles(self).contributors or ())

    @property
    def facet_label(self):
        """Facet label getter"""
        request = query_request()
        if IWfTypedSharedContent.providedBy(self):
            data_type = self.get_data_type()
            if data_type is not None:
                i18n = II18n(data_type)
                return i18n.query_attributes_in_order(('facets_label', 'label'),
                                                      request=request)
        shared_tool = get_parent(self, ISharedTool)
        if shared_tool is not None:
            i18n = II18n(shared_tool)
            return i18n.query_attributes_in_order(('facets_label', 'label', 'title'),
                                                  request=request)
        return None

    @property
    def facet_type_label(self):
        """Facet-type label getter"""
        request = query_request()
        if IWfTypedSharedContent.providedBy(self):
            data_type = self.get_data_type()
            if data_type is not None:
                i18n = II18n(data_type)
                return i18n.query_attributes_in_order(('facets_type_label', 'label'),
                                                      request=request)
        shared_tool = get_parent(self, ISharedTool)
        if shared_tool is not None:
            i18n = II18n(shared_tool)
            return i18n.query_attributes_in_order(('facets_type_label', 'label', 'title'),
                                                  request=request)
        return None

    def elastic_mapping(self):
        """Elasticsearch mapping getter"""
        return IDocumentIndexInfo(self)

    def elastic_document(self):
        document_info = super().elastic_document()
        registry = get_pyramid_registry()
        for name, adapter in registry.getAdapters((self,), IDocumentIndexInfo):
            if not name:
                continue
            document_info.update(adapter)
        return document_info


WfSharedContent.__bases__ += (ElasticDocumentMixin,)
classImplements(WfSharedContent, IDocumentIndexTarget)


@adapter_config(required=IWfSharedContent,
                provides=IDocumentIndexInfo)
def shared_content_index_info(content):
    """Shared content index info"""
    return ESMapping(properties=ESMapping(ESField('@timestamp', attr='timestamp'),
                                          ESKeyword('internal_id'),
                                          ESKeyword('reference_id'),
                                          ESKeyword('owner_id'),
                                          ESKeyword('contributor_id'),
                                          ESKeyword('content_type'),
                                          ESKeyword('facet_label'),
                                          ESKeyword('facet_type_label'),
                                          ESText('title'),
                                          ESText('short_name'),
                                          ESText('header'),
                                          ESText('description'),
                                          ESText('keywords')))


@adapter_config(required=IWfTypedSharedContent,
                provides=IDocumentIndexInfo)
def typed_shared_content_index_info(content):
    """Typed shared content index info"""
    return ESMapping(properties=ESMapping(ESField('@timestamp', attr='timestamp'),
                                          ESKeyword('internal_id'),
                                          ESKeyword('reference_id'),
                                          ESKeyword('owner_id'),
                                          ESKeyword('contributor_id'),
                                          ESKeyword('content_type'),
                                          ESKeyword('data_type'),
                                          ESKeyword('facet_label'),
                                          ESKeyword('facet_type_label'),
                                          ESText('title'),
                                          ESText('short_name'),
                                          ESText('header'),
                                          ESText('description'),
                                          ESText('keywords')))


@adapter_config(name='path',
                required=IWfSharedContent,
                provides=IDocumentIndexInfo)
def shared_content_path_index_info(content):
    """Shared content path index info"""
    return {
        'parent_ids': IPathElements(content).parents
    }


@subscriber(IObjectAddedEvent, context_selector=IDocumentIndexTarget)
def handle_added_document(event):
    """Handle added document"""
    indexer = query_utility(IContentIndexerUtility)
    if indexer is None:
        return
    indexer.index_document(event.object)


INDEXED_DOCUMENTS_EXTENSION = 'pyams_content_es.indexed_documents'


@subscriber(IObjectModifiedEvent, context_selector=IDocumentIndexTarget)
def handle_modified_document(event):
    """Handle modified document

    We use transaction annotations to avoid several indexations of the same
    document during a single transaction...
    """
    indexer = query_utility(IContentIndexerUtility)
    if indexer is None:
        return
    intids = get_utility(IIntIds)
    document = event.object
    document_id = intids.register(document)
    transaction = ITransactionManager(document).get()
    documents = transaction.extension.get(INDEXED_DOCUMENTS_EXTENSION) or set()
    if document_id not in documents:
        documents.add(document_id)
        transaction.extension[INDEXED_DOCUMENTS_EXTENSION] = documents
        indexer.index_document(document)


@subscriber(IObjectRemovedEvent, context_selector=IDocumentIndexTarget)
def handle_removed_document(event):
    """Handle removed document"""
    indexer = query_utility(IContentIndexerUtility)
    if indexer is None:
        return
    document = event.object
    oid = ISequentialIdInfo(document).hex_oid
    wf_state = IWorkflowState(document, None)
    document_oid = f'{oid}.{wf_state.version_id}' if wf_state else oid
    indexer.unindex_document(document_oid)


@subscriber(IObjectAddedEvent)
@subscriber(IObjectModifiedEvent)
@subscriber(IObjectRemovedEvent)
def handle_modified_inner_content(event):
    """Handle modified shared object inner content

    This generic subscriber is used to update index on any content modification.
    """
    source = event.object
    if IWfSharedContent.providedBy(source):
        return
    registry = get_pyramid_registry()
    handler = registry.queryAdapter(source, IPreventSharedContentUpdateSubscribers)
    if handler is not None:
        return
    content = get_parent(event.object, IWfSharedContent, allow_context=False)
    if content is None:
        return
    registry.notify(ObjectModifiedEvent(content))



#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

__docformat__ = 'restructuredtext'

from elasticsearch_dsl import Q
from zope.intid.interfaces import IIntIds

from pyams_content.component.thesaurus import ICollectionsManager, ITagsManager, IThemesManager
from pyams_content.shared.view import IWfView
from pyams_content.shared.view.interfaces.query import IViewQueryParamsExtension
from pyams_content.shared.view.interfaces.settings import IViewCollectionsSettings, IViewTagsSettings, \
    IViewThemesSettings
from pyams_content_es.shared.view import IEsViewUserQuery
from pyams_thesaurus.interfaces.thesaurus import IThesaurus
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.registry import get_utility


#
# Tags queries
#

@adapter_config(name='tags',
                required=IWfView,
                provides=IViewQueryParamsExtension)
class EsViewTagsQueryParamsExtension(ContextAdapter):
    """Elasticsearch view tags query params extension"""

    weight = 50

    def get_params(self, context, request=None):
        """Query params getter"""
        settings = IViewTagsSettings(self.context)
        tags = settings.get_tags_index(context)
        if tags:
            yield Q('terms',
                    **{'tags': tags})
        elif settings.select_context_tags:
            yield None


@adapter_config(name='tags',
                required=IWfView,
                provides=IEsViewUserQuery)
class EsViewTagsUserQuery(ContextAdapter):
    """Elasticsearch view tags user params"""

    weight = 50

    @staticmethod
    def get_user_params(request):
        """Query user params getter"""
        tags = request.params.getall('tags')
        if not tags:
            return
        manager = ITagsManager(request.root, None)
        if manager is None:
            return
        thesaurus = get_utility(IThesaurus, name=manager.thesaurus_name)
        if thesaurus is None:
            return
        if isinstance(tags, str):
            tags = tags.split(',')
        intids = get_utility(IIntIds)
        yield Q('terms',
                **{'tags': [
                    intids.queryId(term)
                    for term in [
                        thesaurus.terms.get(value)
                        for tag in tags
                        for value in tag.split(',')
                    ]
                    if term is not None
                ]})


#
# Themes queries
#

@adapter_config(name='themes',
                required=IWfView,
                provides=IViewQueryParamsExtension)
class EsViewThemesQueryParamsExtension(ContextAdapter):
    """Elasticsearch view themes query params extension"""

    weight = 52

    def get_params(self, context, request=None):
        """Query params getter"""
        settings = IViewThemesSettings(self.context)
        tags = settings.get_themes_index(context)
        if tags:
            yield Q('terms',
                    **{'themes.terms': tags})
        elif settings.select_context_themes:
            yield None


@adapter_config(name='themes',
                required=IWfView,
                provides=IEsViewUserQuery)
class EsViewThemesUserQuery(ContextAdapter):
    """Elasticsearch view themes user params"""

    weight = 52

    @staticmethod
    def get_user_params(request):
        """Query user params getter"""
        themes = request.params.get('themes')
        if not themes:
            return
        manager = IThemesManager(request.root, None)
        if manager is None:
            return
        thesaurus = get_utility(IThesaurus, name=manager.thesaurus_name)
        if thesaurus is None:
            return
        if isinstance(themes, str):
            themes = themes.split(',')
        intids = get_utility(IIntIds)
        yield Q('terms',
                **{'themes.terms': [
                    intids.queryId(term)
                    for term in [
                        thesaurus.terms.get(value)
                        for tag in themes
                        for value in tag.split(',')
                    ]
                    if term is not None
                ]})


#
# Collections queries
#

@adapter_config(name='collections',
                required=IWfView,
                provides=IViewQueryParamsExtension)
class EsViewCollectionsQueryParamsExtension(ContextAdapter):
    """Elasticsearch view collections query params extension"""

    weight = 54

    def get_params(self, context, request=None):
        """Query params getter"""
        settings = IViewCollectionsSettings(self.context)
        tags = settings.get_collections_index(context)
        if tags:
            yield Q('terms',
                    **{'collections': tags})
        elif settings.select_context_collections:
            yield None


@adapter_config(name='collections',
                required=IWfView,
                provides=IEsViewUserQuery)
class EsViewCollectionsUserQuery(ContextAdapter):
    """Elasticsearch view collections user params"""

    weight = 54

    @staticmethod
    def get_user_params(request):
        """Query user params getter"""
        collections = request.params.get('collections')
        if not collections:
            return
        manager = ICollectionsManager(request.root, None)
        if manager is None:
            return
        thesaurus = get_utility(IThesaurus, name=manager.thesaurus_name)
        if thesaurus is None:
            return
        if isinstance(collections, str):
            collections = collections.split(',')
        intids = get_utility(IIntIds)
        yield Q('terms',
                **{'collections': [
                    intids.queryId(term)
                    for term in [
                        thesaurus.terms.get(value)
                        for tag in collections
                        for value in tag.split(',')
                    ]
                    if term is not None
                ]})



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.component.links module

This module defines the adapters which are used to handle links indexation.
"""

from pyams_content.component.association.interfaces import IAssociationContainer, IAssociationItem
from pyams_content.component.links import IBaseLink
from pyams_content.component.paragraph.interfaces import IParagraphContainer, \
    IParagraphContainerTarget
from pyams_content_es.interfaces import IDocumentIndexInfo
from pyams_utils.adapter import adapter_config


__docformat__ = 'restructuredtext'


@adapter_config(name='links',
                required=IParagraphContainerTarget,
                provides=IDocumentIndexInfo)
def paragraph_container_target_link_index_info(context):
    """Internal and external links index info"""
    links = []
    for paragraph in IParagraphContainer(context).get_visible_paragraphs():
        associations = IAssociationContainer(paragraph, {})
        for link in associations.values():
            if not IAssociationItem(link).visible:
                continue
            if not IBaseLink.providedBy(link):
                continue
            links.append({
                'title': link.title,
                'description': link.description
            })
    return {
        'link': links
    }



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS content ES.include module

This module is used for Pyramid integration
"""
import atexit
import os
import re
import sys
from pyramid.events import subscriber
from pyramid.interfaces import IApplicationCreated
from pyramid.settings import asbool

from pyams_content_es.interfaces import INDEXER_AUTH_KEY, INDEXER_CLIENTS_KEY, \
    INDEXER_HANDLER_KEY, INDEXER_NAME, \
    INDEXER_STARTER_KEY, \
    LOGGER
from pyams_content_es.process import ContentIndexerMessageHandler, ContentIndexerProcess
from pyams_site.interfaces import PYAMS_APPLICATION_DEFAULT_NAME, PYAMS_APPLICATION_SETTINGS_KEY
from pyams_utils.protocol.tcp import is_port_in_use
from pyams_utils.registry import get_pyramid_registry, set_local_registry
from pyams_utils.zodb import get_connection_from_settings
from pyams_zmq.process import process_exit_func


__docformat__ = 'restructuredtext'


def include_package(config):
    """Pyramid package include"""

    # add translations
    config.add_translation_dirs('pyams_content_es:locales')

    # override PyAMS_content package components
    config.include('pyams_content')

    try:
        import pyams_zmi  # pylint: disable=import-outside-toplevel,unused-import
    except ImportError:
        config.scan(ignore=[re.compile(r'pyams_content_es\..*\.zmi\.?.*').search])
    else:
        config.scan()


@subscriber(IApplicationCreated)
def handle_new_application(event):
    """Start Elasticsearch document indexer when application is created"""

    # Check for PyAMS command line script
    cmdline = os.path.split(sys.argv[0])[-1]
    if cmdline.startswith('pyams_'):
        return

    registry = get_pyramid_registry()
    settings = registry.settings
    start_handler = asbool(settings.get(INDEXER_STARTER_KEY, False))
    if not start_handler:
        return

    # check if port is available
    handler_address = settings.get(INDEXER_HANDLER_KEY, '127.0.0.1:5557')
    hostname, port = handler_address.split(':')
    if is_port_in_use(int(port), hostname):
        LOGGER.warning("Elasticsearch indexer port already used, aborting...")
        return

    # get database connection
    connection = get_connection_from_settings(settings)
    root = connection.root()
    # get application
    application_name = settings.get(PYAMS_APPLICATION_SETTINGS_KEY,
                                    PYAMS_APPLICATION_DEFAULT_NAME)
    application = root.get(application_name)
    if application is None:
        return

    process = None
    sm = application.getSiteManager()
    set_local_registry(sm)
    try:
        indexer = sm.get(INDEXER_NAME)
        if indexer is None:
            return
        # create indexer process
        process = ContentIndexerProcess(handler_address,
                                        ContentIndexerMessageHandler,
                                        settings.get(INDEXER_AUTH_KEY),
                                        settings.get(INDEXER_CLIENTS_KEY),
                                        registry)
        LOGGER.info(f"Starting Elasticsearch documents indexer {process!r}...")
        process.start()
        if process.is_alive():
            atexit.register(process_exit_func, process=process)
            LOGGER.info(f"Started Elasticsearch documents indexer with PID {process.pid}.")
    finally:
        if process and not process.is_alive():
            process.terminate()
            process.join()
        set_local_registry(None)



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.component.gallery module

This module index content extensions for media galleries.
"""

from pyams_content.component.gallery import IBaseGallery, IGallery, IGalleryTarget
from pyams_content_es.component import get_index_values
from pyams_content_es.interfaces import IDocumentIndexInfo
from pyams_utils.adapter import adapter_config

__docformat__ = 'restructuredtext'


@adapter_config(context=IBaseGallery,
                provides=IDocumentIndexInfo)
def gallery_index_info(gallery):
    """Gallery index info"""
    info = {}
    get_index_values(gallery, info,
                     i18n_fields=('title', 'description'))
    for image in gallery.get_visible_images():
        get_index_values(image, info,
                         fields=('author',),
                         i18n_fields=('title', 'description', 'author_comments',
                                      'sound_title', 'sound_description'))
    return info


@adapter_config(name='gallery',
                required=IGalleryTarget,
                provides=IDocumentIndexInfo)
def gallery_target_index_info(context):
    """Gallery target index info"""
    gallery = IGallery(context)
    info = IDocumentIndexInfo(gallery, None)
    return {
        'gallery': info
    }



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.component.paragraph module

This module defines the adapters which are used to handle paragraphs indexation.
"""

__docformat__ = 'restructuredtext'

from pyams_content.component.paragraph.interfaces import IBaseParagraph, IParagraphContainer, \
    IParagraphContainerTarget
from pyams_content.component.paragraph.interfaces.group import IParagraphsGroup
from pyams_content.component.paragraph.interfaces.html import IHTMLParagraph, IRawParagraph
from pyams_content_es.component import get_index_values, html_to_index
from pyams_content_es.interfaces import IDocumentIndexInfo
from pyams_utils.adapter import adapter_config


@adapter_config(name='paragraphs',
                required=IParagraphContainerTarget,
                provides=IDocumentIndexInfo)
def paragraph_container_index_info(context):
    """Paragraph container index info"""
    body = {}
    for paragraph in IParagraphContainer(context).get_visible_paragraphs():
        info = IDocumentIndexInfo(paragraph, None)
        if info is not None:
            for lang, body_info in info.items():
                body[lang] = f"{body.get(lang, '')}\n{body_info}"
    return {
        'body': body
    }


@adapter_config(required=IBaseParagraph,
                provides=IDocumentIndexInfo)
def base_paragraph_index_info(context):
    """Base paragraph index info"""
    info = {}
    get_index_values(context, info,
                     i18n_fields=('title',))
    return info


@adapter_config(required=IParagraphsGroup,
                provides=IDocumentIndexInfo)
def paragraphs_group_index_info(context):
    """Paragraphs group index info"""
    info = base_paragraph_index_info(context)
    paragraphs_info = paragraph_container_index_info(context)
    for lang, body in paragraphs_info.get('body', {}).items():
        info[lang] += f"\n{body}"
    return info


@adapter_config(required=IRawParagraph,
                provides=IDocumentIndexInfo)
@adapter_config(required=IHTMLParagraph,
                provides=IDocumentIndexInfo)
def html_paragraph_index_info(context):
    """HTML paragraph index info"""
    info = base_paragraph_index_info(context)
    get_index_values(context, info,
                     i18n_fields=(('body', html_to_index),))
    return info



#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.shared.view.reference module

"""

__docformat__ = 'restructuredtext'

from elasticsearch_dsl import Q
from zope.intid.interfaces import IIntIds

from pyams_content.shared.view import IWfView
from pyams_content.shared.view.interfaces.query import EXCLUDED_VIEW_ITEMS
from pyams_content.shared.view.interfaces.settings import IViewInternalReferencesSettings, ONLY_REFERENCE_MODE
from pyams_content_es.shared.view import IEsViewQueryParamsExtension, IEsViewUserQuery
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.registry import get_utility


@adapter_config(name='references',
                required=IWfView,
                provides=IEsViewQueryParamsExtension)
class EsViewReferencesQueryParamsExtension(ContextAdapter):
    """Elasticsearch view internal references query params extension"""

    weight = 10

    def get_params(self, context, request=None):
        """Elasticsearch query params getter"""
        settings = IViewInternalReferencesSettings(self.context)
        if settings.references_mode == ONLY_REFERENCE_MODE:
            yield None
        else:
            if settings.exclude_context:
                intids = get_utility(IIntIds)
                try:
                    yield Q('bool',
                            must_not=Q('term',
                                       **{'internal_id': intids.register(context)}))
                except TypeError:
                    return


@adapter_config(name='exclusions',
                required=IWfView,
                provides=IEsViewUserQuery)
class EsExclusionsViewQueryParamsExtension(ContextAdapter):
    """Elasticsearch exclusions for Elasticsearch

    This adapter is looking into request's annotations for items which should be excluded
    from search.
    """

    @staticmethod
    def get_user_params(request):
        excluded_items = request.annotations.get(EXCLUDED_VIEW_ITEMS)
        if excluded_items:
            yield Q('bool',
                    must_not=Q('terms',
                               **{'reference_id': tuple(excluded_items)}))



#
# Copyright (c) 2015-2024 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

import locale

from zope.intid.interfaces import IIntIds

from pyams_content.feature.filter import ICollectionsFilter, IContentTypesFilter, IFilter, ITagsFilter, IThemesFilter, \
    ITitleFilter
from pyams_content.feature.filter.interfaces import FILTER_SORTING, IAggregatedPortletRendererSettings, IFilterProcessor
from pyams_content.feature.filter.processor import BaseFilterProcessor
from pyams_layer.interfaces import IPyAMSLayer
from pyams_utils.adapter import adapter_config
from pyams_utils.registry import get_utility

__docformat__ = 'restructuredtext'


SORTING_MODES = {
    FILTER_SORTING.ALPHA.value: lambda x: locale.strxfrm(x['label']),
    FILTER_SORTING.ALPHA_DESC.value: lambda x: locale.strxfrm(x['label']),
    FILTER_SORTING.COUNT.value: lambda x: (x['doc_count'], locale.strxfrm(x['label'])),
    FILTER_SORTING.COUNT_DESC.value: lambda x: (x['doc_count'], locale.strxfrm(x['label'])),
    FILTER_SORTING.MANUAL.value: lambda x: x.get('order', 0)
}


def get_sorting_key(sorting_mode):
    return SORTING_MODES.get(sorting_mode, 'alpha_asc')


@adapter_config(required=(IFilter, IPyAMSLayer, IAggregatedPortletRendererSettings),
                provides=IFilterProcessor)
class EsBaseFilterProcessor(BaseFilterProcessor):
    """Elasticsearch base filter processor"""

    def get_aggregations(self, aggregations):
        result = self.convert_aggregations(aggregations)
        sorting_mode = self.filter.sorting_mode
        sorting_key = get_sorting_key(sorting_mode)
        if sorting_key:
            result.sort(key=sorting_key,
                        reverse=sorting_mode.endswith('_desc'))
        return result

    def convert_aggregations(self, aggregations):
        result = []
        for item in aggregations:
            result.append({
                'key': item.key,
                'label': item.key,
                'doc_count': item.doc_count
            })
        return result


@adapter_config(required=(IContentTypesFilter, IPyAMSLayer, IAggregatedPortletRendererSettings),
                provides=IFilterProcessor)
class ContentTypeFilterProcessor(EsBaseFilterProcessor):
    """Content-type filter processor"""

    def process(self, aggregations, filter_type=None):
        return super().process(aggregations, self.filter.content_mode)


@adapter_config(required=(ITitleFilter, IPyAMSLayer, IAggregatedPortletRendererSettings),
                provides=IFilterProcessor)
class TitleFilterProcessor(EsBaseFilterProcessor):
    """Title filter processor"""


class EsBaseThesaurusFilterProcessor(EsBaseFilterProcessor):
    """Base filter aggregation processor"""

    def convert_aggregations(self, aggregations):
        """Convert aggregation raw data in human form"""
        intids = get_utility(IIntIds)
        result = []
        for item in aggregations:
            tag_id = int(item.key)
            tag_object = intids.queryObject(tag_id)
            if tag_object:
                result.append({
                    'key': tag_object.label,
                    'label': tag_object.public_title,
                    'order': tag_object.order or 0,
                    'doc_count': item.doc_count
                })
        return result


@adapter_config(required=(ITagsFilter, IPyAMSLayer, IAggregatedPortletRendererSettings),
                provides=IFilterProcessor)
class TagsFilterProcessor(EsBaseThesaurusFilterProcessor):
    """Tags filter processor"""


@adapter_config(required=(IThemesFilter, IPyAMSLayer, IAggregatedPortletRendererSettings),
                provides=IFilterProcessor)
class ThemeFilterProcessor(EsBaseThesaurusFilterProcessor):
    """Themes filter processor"""


@adapter_config(required=(ICollectionsFilter, IPyAMSLayer, IAggregatedPortletRendererSettings),
                provides=IFilterProcessor)
class CollectionFilterProcessor(EsBaseThesaurusFilterProcessor):
    """Collections filter processor"""



#
# Copyright (c) 2015-2022 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_content_es.scripts module

This module defines a script entry point which can be used to reindex all site contents.
"""

__docformat__ = 'restructuredtext'

import argparse
import sys
import textwrap
from pyramid.paster import bootstrap

from pyams_content_es.utils import index_site


def pyams_index_cmd():
    """Update Elasticsearch index with all contents"""
    usage = f"usage: {sys.argv[0]} config_uri [-i content_type]* [-x content_type]*"
    description = """Update Elasticsearch index with all database contents."""
    parser = argparse.ArgumentParser(usage=usage,
                                     description=textwrap.dedent(description))
    parser.add_argument('-c', '--check', action='store_true',
                        help="If set, only reindex missing documents")
    parser.add_argument('-r', '--root',
                        help="Define path of indexation root")
    parser.add_argument('-i', '--include', action='append',
                        help="Included shared content types; can be 'topic', 'event'... "
                             "You can set this argument several times to specify several "
                             "content types to include into index")
    parser.add_argument('-x', '--exclude', action='append',
                        help="Excluded shared content types; can be 'topic', 'event'... "
                             "You can set this argument several times to specify several "
                             "content types to exclude from index")
    parser.add_argument('-t', '--timeout', type=float, default=10,
                        help="Document indexer timeout, in seconds (default=10)")
    parser.add_argument('-v', '--verbose', action='store_true',
                        help="Verbose outputs")
    parser.add_argument('config_uri', help="Configuration filename")
    args = parser.parse_args()

    if args.include and args.exclude:
        print("INCLUDE and EXCLUDE arguments can't be used simultaneously!")
        sys.exit(1)

    env = bootstrap(args.config_uri)
    closer = env['closer']
    try:
        index_site(env['request'], cmd_args=args)
    finally:
        closer()
